import copy
import pathlib
from typing import Any, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize, to_hex

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils import Lineage
from cellrank._utils._docs import d
from cellrank._utils._utils import _unique_order_preserving, save_fig
from cellrank.pl._utils import _get_categorical_colors, _position_legend

__all__ = ["log_odds"]


@d.dedent
def log_odds(
    adata: AnnData,
    lineage_1: str,
    lineage_2: Optional[str] = None,
    time_key: str = "exp_time",
    backward: bool = False,
    keys: Optional[Union[str, Sequence[str]]] = None,
    threshold: Optional[Union[float, Sequence]] = None,
    threshold_color: str = "red",
    layer: Optional[str] = None,
    use_raw: bool = False,
    size: float = 2.0,
    cmap: str = "viridis",
    alpha: Optional[float] = 0.8,
    ncols: Optional[int] = None,
    fontsize: Optional[Union[float, str]] = None,
    xticks_step_size: Optional[int] = 1,
    legend_loc: Optional[str] = "best",
    jitter: Union[bool, float] = True,
    seed: Optional[int] = None,
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[int] = None,
    save: Optional[Union[str, pathlib.Path]] = None,
    show: bool = True,
    **kwargs: Any,
) -> Optional[Union[Axes, Sequence[Axes]]]:
    """Plot log-odds ratio between trajectories.

    This plotting function is geared towards time-series datasets that have been analyzed
    using the :class:`~cellrank.kernels.RealTimeKernel`. It visualizes log-odd ratios
    between different trajectories per cell, with the option to color in certain
    molecular features, like genes. This can be useful to detect and visualize fate-decisive genes.

    Parameters
    ----------
    %(adata)s
    lineage_1
        The first lineage for which to compute the log-odds.
    lineage_2
        The second lineage for which to compute the log-odds. If :obj:`None`, use the rest of the lineages.
    time_key
        Key in :attr:`~anndata.AnnData.obs` containing the experimental time.
    %(backward)s
    keys
        Key in :attr:`~anndata.AnnData.obs` or :attr:`~anndata.AnnData.var_names`.
    threshold
        Visualize whether total expression per cell is greater than ``threshold``.
        If a :class:`~typing.Sequence`, it should be the same length as ``keys``.
    threshold_color
        Color to use when plotting thresholded expression values.
    layer
        Which layer to use to get expression values. If :obj:`None` or ``'X'``, use :attr:`~anndata.AnnData.X`.
    use_raw
        Whether to access :attr:`~anndata.AnnData.raw`. If :obj:`True`, ``layer`` is ignored.
    size
        Size of the dots.
    cmap
        Colormap to use for continuous variables in ``keys``.
    alpha
        Alpha values for the dots.
    ncols
        Number of columns.
    fontsize
        Size of the font for the title, x- and y-label.
    xticks_step_size
        Show only every other *n-th* tick on the x-axis. If :obj:`None`, don't show any ticks.
    legend_loc
        Position of the legend. If :obj:`None`, do not show the legend.
    jitter
        Amount of jitter to apply along x-axis.
    seed
        Seed for ``jitter`` to ensure reproducibility.
    %(plotting)s
    show
        If `False`, return :class:`~matplotlib.axes.Axes` or a sequence of them.
    kwargs
        Keyword arguments for :func:`~seaborn.stripplot`.

    Returns
    -------
    %(just_plots)s
    If ``show = False``, returns the axes object.
    """
    from cellrank.kernels._utils import _ensure_numeric_ordered

    def decorate(ax: Axes, *, title: Optional[str] = None, show_ylabel: bool = True) -> None:
        ax.set_xlabel(time_key, fontsize=fontsize)
        ax.set_title(title, fontdict={"fontsize": fontsize})
        ax.set_ylabel(ylabel if show_ylabel else "", fontsize=fontsize)

        if xticks_step_size is None:
            ax.set_xticks([])
        else:
            step = max(1, xticks_step_size)
            ax.set_xticks(np.arange(0, n_cats, step))
            ax.set_xticklabels(df[time_key].cat.categories[::step])

    def cont_palette(values: np.ndarray) -> Tuple[List[str], ScalarMappable]:
        cm = copy.copy(plt.get_cmap(cmap))
        cm.set_bad("grey")
        sm = ScalarMappable(cmap=cm, norm=Normalize(vmin=np.nanmin(values), vmax=np.nanmax(values)))
        return [to_hex(v) for v in (sm.to_rgba(values))], sm

    def get_data(
        key: str,
        thresh: Optional[float] = None,
    ) -> Tuple[Optional[str], Optional[np.ndarray], Optional[np.ndarray], ScalarMappable]:
        try:
            _, palette = _get_categorical_colors(adata, key)
            df[key] = adata.obs[key].values[mask]
            df[key] = df[key].cat.remove_unused_categories()
            try:
                # seaborn doesn't like numeric categories
                df[key] = df[key].astype(float)
                palette = {float(k): v for k, v in palette.items()}
            except ValueError:
                pass
            # otherwise seaborn plots all
            palette = {k: palette[k] for k in df[key].unique()}
            hue, thresh_mask, sm = key, None, None
        except TypeError:
            # not a categorical
            palette, hue, thresh_mask, sm = (
                cont_palette(adata.obs[key].values[mask])[0],
                None,
                None,
                None,
            )
        except KeyError:
            # unable to find data in `adata.obs`, try res
            try:
                # fmt: off
                if thresh is None:
                    values = adata.raw.obs_vector(key) if use_raw else adata.obs_vector(key, layer=layer)
                    palette, sm = cont_palette(values)
                    hue, thresh_mask = None, None
                else:
                    if use_raw:
                        values = np.asarray(adata.raw[:, key].X[mask].sum(1)).squeeze()
                    elif layer not in (None, "X"):
                        values = np.asarray(adata[:, key].layers[layer][mask].sum(1)).squeeze()
                    else:
                        values = np.asarray(adata[:, key].X[mask].sum(1)).squeeze()
                    thresh_mask = values > thresh
                    hue, palette, sm = None, None, None
                # fmt: on
            except KeyError as e:
                raise e from None

        return hue, palette, thresh_mask, sm

    np.random.seed(seed)  # noqa: NPY002
    _ = kwargs.pop("orient", None)
    if use_raw and adata.raw is None:
        logg.warning("No raw attribute set. Setting `use_raw=False`")
        use_raw = False

    probs = Lineage.from_adata(adata, backward=backward)
    time = _ensure_numeric_ordered(adata, time_key)
    order = time.cat.categories[:: -1 if backward else 1]

    fate1 = probs[lineage_1].X.squeeze(-1)
    if lineage_2 is None:
        fate2 = 1 - fate1
        ylabel = rf"$\log{{\frac{{{lineage_1}}}{{rest}}}}$"
    else:
        fate2 = probs[lineage_2].X.squeeze(-1)
        ylabel = rf"$\log{{\frac{{{lineage_1}}}{{{lineage_2}}}}}$"

    # fmt: off
    df = pd.DataFrame(
        {
            "log_odds": np.log(np.divide(fate1, fate2, where=fate2 != 0, out=np.zeros_like(fate1)) + 1e-12),
            time_key: time,
        }
    )
    mask = (fate1 != 0) & (fate2 != 0)
    df = df[mask]
    n_cats = len(df[time_key].cat.categories)
    # fmt: on

    if keys is None:
        if figsize is None:
            figsize = np.array([n_cats, n_cats * 4 / 6]) / 2

        fig, ax = plt.subplots(figsize=figsize, dpi=dpi, tight_layout=True)
        ax = sns.stripplot(
            x=time_key,
            y="log_odds",
            data=df,
            order=order,
            jitter=jitter,
            color="k",
            size=size,
            ax=ax,
            **kwargs,
        )
        decorate(ax)
        if save is not None:
            save_fig(fig, save)
        return None if show else ax

    if isinstance(keys, str):
        keys = (keys,)
    if not len(keys):
        raise ValueError("No keys have been selected.")
    keys = _unique_order_preserving(keys)

    if not isinstance(threshold, Iterable):
        threshold = (threshold,) * len(keys)
    if len(threshold) != len(keys):
        raise ValueError(f"Expected `threshold` to be of length `{len(keys)}`, found `{len(threshold)}`.")

    ncols = max(len(keys) if ncols is None else ncols, 1)
    nrows = int(np.ceil(len(keys) / ncols))
    if figsize is None:
        figsize = np.array([n_cats * ncols, n_cats * nrows * 4 / 6]) / 2

    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        figsize=figsize,
        dpi=dpi,
        constrained_layout=True,
        sharey="all",
    )
    axes = np.ravel([axes])

    i = 0
    for i, (key, ax, thresh) in enumerate(zip(keys, axes, threshold)):
        hue, palette, thresh_mask, sm = get_data(key, thresh)
        show_ylabel = i % ncols == 0

        ax = sns.stripplot(
            x=time_key,
            y="log_odds",
            data=df if thresh_mask is None else df[~thresh_mask],
            hue=hue,
            order=order,
            jitter=jitter,
            color="black",
            palette=palette,
            size=size,
            alpha=alpha if alpha is not None else None if thresh_mask is None else 0.8,
            ax=ax,
            **kwargs,
        )
        if thresh_mask is not None:
            sns.stripplot(
                x=time_key,
                y="log_odds",
                data=df if thresh_mask is None else df[thresh_mask],
                hue=hue,
                order=order,
                jitter=jitter,
                color=threshold_color,
                palette=palette,
                size=size * 2,
                alpha=0.9,
                ax=ax,
                **kwargs,
            )
            key = rf"${key} > {thresh}$"
        if sm is not None:
            cax = ax.inset_axes([1.02, 0, 0.025, 1], transform=ax.transAxes)
            fig.colorbar(sm, ax=ax, cax=cax)
        else:
            if legend_loc in (None, "none"):
                legend = ax.get_legend()
                if legend is not None:
                    legend.remove()
            else:
                handles, labels = ax.get_legend_handles_labels()
                if len(handles):
                    _position_legend(ax, legend_loc=legend_loc, handles=handles, labels=labels)

        decorate(ax, title=key, show_ylabel=show_ylabel)

    for ax in axes[i + 1 :]:
        ax.remove()
    axes = axes[: i + 1]

    if save is not None:
        save_fig(fig, save)

    return None if show else axes[0] if len(axes) == 1 else axes



from typing import Any, Optional, Union

import numpy as np
import scipy.sparse as sp

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._key import Key
from cellrank._utils._utils import _read_graph_data
from cellrank.kernels._base_kernel import KernelExpression, UnidirectionalKernel

__all__ = ["PrecomputedKernel"]


@d.dedent
class PrecomputedKernel(UnidirectionalKernel):
    """Kernel which is initialized based on a precomputed transition matrix.

    This kernel serves as CellRank's interface with other methods that compute cell-cell transition matrices; you
    can use this kernel to input you own custom transition matrix and continue to use all CellRank functionality.
    In particular, you can use a precomputed kernel, just like any other kernel, to initialize an estimator and
    compute initial and terminal states, fate probabilities, and driver genes.

    Parameters
    ----------
    object
        Can be one of the following types:

        - :class:`~anndata.AnnData` - annotated data object.
        - :class:`~numpy.ndarray`, :class:`~scipy.sparse.spmatrix` - row-normalized transition matrix.
        - :class:`~cellrank.kernels.Kernel` - kernel.
        - :class:`str` - key in :attr:`~anndata.AnnData.obsp` where the transition matrix is stored.
          ``adata`` must be provided in this case.
        - :class:`bool` - directionality of the transition matrix that will be used to infer its storage location.
          If :obj:`None`, the directionality will be determined automatically. ``adata`` must be provided in this case.
    %(adata)s
        Must be provided when ``object`` is :class:`str` or :class:`bool`.
    obsp_key
        Key in :attr:`~anndata.AnnData.obsp` where the transition matrix is stored.
        If :obj:`None`, it will be determined automatically. Only used when ``object`` is :class:`~anndata.AnnData`.
    copy
        Whether to copy the stored transition matrix.
    backward
        Hint whether this is a forward, backward or a unidirectional kernel.
        Only used when ``object`` is :class:`~anndata.AnnData`.

    Notes
    -----
    If ``object`` is :class:`~anndata.AnnData` and neither ``obsp_key`` nor ``backward`` is specified,
    default forward and backward are tried and first one is used.
    """

    _SENTINEL = object()

    def __init__(
        self,
        object: Union[str, bool, np.ndarray, sp.spmatrix, AnnData, KernelExpression],
        adata: Optional[AnnData] = None,
        obsp_key: Optional[str] = None,
        **kwargs: Any,
    ):
        if isinstance(object, AnnData):
            self._from_adata(object, obsp_key=obsp_key, **kwargs)
        elif isinstance(object, KernelExpression):
            self._from_kernel(object, copy=kwargs.get("copy", False))
        elif isinstance(object, (np.ndarray, sp.spmatrix)):
            self._from_matrix(object, adata=adata, **kwargs)
        elif isinstance(adata, AnnData):
            if isinstance(object, str):
                self._from_adata(adata, obsp_key=object, **kwargs)
            elif object is None or isinstance(object, bool):
                kwargs["backward"] = object
                self._from_adata(adata, **kwargs)
            else:
                raise ValueError("Unable to interpret the data.")
        else:
            raise TypeError(
                f"Expected `object` to be either `str`, `bool`, `numpy.ndarray`, "
                f"`scipy.sparse.spmatrix`, `AnnData` or `KernelExpression`, "
                f"found `{type(object).__name__}`."
            )

    def _from_adata(
        self,
        adata: AnnData,
        obsp_key: Optional[str] = None,
        backward: Optional[bool] = _SENTINEL,
        copy: bool = False,
    ) -> None:
        if obsp_key is None:
            if backward is PrecomputedKernel._SENTINEL:
                for backward in [False, True]:  # prefer `forward`
                    obsp_key = Key.uns.kernel(backward)
                    if obsp_key in adata.obsp:
                        break
                else:
                    raise ValueError(
                        "Unable to find transition matrix in `adata.obsp`. "
                        "Please specify `obsp_key=...` or `backward=...`."
                    )
            else:
                obsp_key = Key.uns.kernel(backward)
            logg.info(f"Using transition matrix from `adata.obsp[{obsp_key!r}]`")

        tmat = _read_graph_data(adata, obsp_key)

        if backward is PrecomputedKernel._SENTINEL:
            # not ideal, since None/False share the same key, prefer `forward`
            if obsp_key == Key.uns.kernel(bwd=False):
                backward = False
            elif obsp_key == Key.uns.kernel(bwd=True):
                backward = True
            else:
                backward = None
            logg.info(f"Setting directionality `backward={backward}`")

        self._from_matrix(tmat, adata=adata, backward=backward, copy=copy)
        self.params["origin"] = f"adata.obsp[{obsp_key!r}]"
        self._init_kwargs = {"obsp_key": obsp_key, "backward": backward}

    def _from_kernel(self, kernel: KernelExpression, copy: bool = False) -> None:
        if kernel.transition_matrix is None:
            raise RuntimeError("Compute transition matrix first as `.compute_transition_matrix()`.")
        self._from_matrix(
            kernel.transition_matrix,
            backward=kernel.backward,
            adata=kernel.adata,
            copy=copy,
        )
        self._params = kernel.params.copy()
        self.params["origin"] = repr(kernel)

    def _from_matrix(
        self,
        matrix: Union[np.ndarray, sp.spmatrix],
        adata: Optional[AnnData] = None,
        backward: Optional[bool] = None,
        copy: bool = False,
    ) -> None:
        # fmt: off
        if adata is None:
            logg.warning(f"Creating empty `AnnData` object of shape `{matrix.shape[0], 1}`")
            adata = AnnData(sp.csr_matrix((matrix.shape[0], 1)))
        super().__init__(adata)
        self._backward: Optional[bool] = backward
        self.transition_matrix = matrix.copy() if copy else matrix
        self.params["origin"] = "array"
        # fmt: on

    def compute_transition_matrix(self, *_: Any, **__: Any) -> "PrecomputedKernel":
        """Do nothing and return self."""
        return self

    @property
    def backward(self) -> Optional[bool]:
        """Direction of the process."""
        return self._backward



from importlib import metadata

from cellrank import datasets, estimators, kernels, logging, models, pl
from cellrank._utils._lineage import Lineage
from cellrank.settings import settings

__all__ = ["datasets", "estimators", "kernels", "logging", "models", "pl", "Lineage", "settings"]

try:
    md = metadata.metadata(__name__)
    __version__ = md.get("version", "")
    __author__ = md.get("Author", "")
    __maintainer__ = md.get("Maintainer-email", "")
except ImportError:
    md = None

del metadata, md



import functools
from typing import List, Optional, Tuple, TypeVar, Union

import numpy as np
import scipy.sparse as sp
from scipy.linalg import solve

from cellrank import logging as logg
from cellrank._utils._enum import DEFAULT_BACKEND
from cellrank._utils._parallelize import _get_n_cores, parallelize

__all__ = ["_solve_lin_system", "_is_petsc_slepc_available"]

_DEFAULT_SOLVER = "gmres"
_PETSC_ERROR_MSG_SHOWN = False
_PETSC_ERROR_MSG = (
    "Unable to import petsc4py. "
    "For installation, please refer to: https://petsc4py.readthedocs.io/en/stable/install.html.\n"
    "Defaulting to `{!r}` solver."
)
_AVAIL_ITER_SOLVERS = {
    "gmres": sp.linalg.gmres,
    "lgmres": sp.linalg.lgmres,
    "bicgstab": sp.linalg.bicgstab,
    "gcrotmk": sp.linalg.gcrotmk,
}

LinSolver = TypeVar("LinSolver")
PETScMat = TypeVar("PETScMat")
Queue = TypeVar("Queue")


def _create_petsc_matrix(mat: Union[np.ndarray, sp.spmatrix], as_dense: bool = False) -> PETScMat:
    """Create a PETSc matrix from :mod:`numpy` or :mod:`scipy.sparse` matrix.

    Parameters
    ----------
    mat
        The matrix to convert.
    as_dense
        Only used when ``mat`` is a sparse matrix. If `True`, create `'seqdense'` matrix,
        otherwise `'seqaij'` matrix.

    Returns
    -------
    The converted matrix.
    """
    # TODO(michalk8): for some solvers, we need to set the diagonal entries explicitly, even if they are zeros
    # see: https://lists.mcs.anl.gov/mailman/htdig/petsc-users/2014-October/023212.html

    from petsc4py import PETSc

    if sp.issparse(mat) and as_dense:
        mat = mat.toarray()

    A = PETSc.Mat().create()
    if sp.issparse(mat):
        if not sp.isspmatrix_csr(mat):
            mat = sp.csr_matrix(mat)
        A.createAIJ(size=mat.shape, csr=(mat.indptr, mat.indices, mat.data))
    else:
        A.createDense(mat.shape, array=mat)

    A.assemblyBegin()
    A.assemblyEnd()

    return A


def _create_solver(
    mat_a: Union[np.ndarray, sp.spmatrix],
    solver: Optional[str],
    preconditioner: Optional[str],
    tol: float,
) -> Tuple["petsc4py.PETSc.KSP", "petsc4py.PETSc.Vec", "petsc4py.PETScVec"]:  # noqa
    """
    Create a linear system solver.

    Parameters
    ----------
    mat_a
        Matrix defining the linear system.
    solver
        Type of the solver. If `None`, use `'gmres'`.
    preconditioner
        Preconditioner to use. If `None`, don't use any.
    tol
        The relative convergence tolerance, relative decrease in the (possibly preconditioned) residual norm .

    Returns
    -------
    Triple containing the solver, vector ``x`` and vector ``b`` in ``A * x = b``.
    """
    from petsc4py import PETSc

    A = _create_petsc_matrix(mat_a)

    ksp = PETSc.KSP().create()
    ksp.setTolerances(rtol=tol)
    ksp.setType(solver if solver is not None else PETSc.KSP.Type.GMRES)

    if preconditioner is not None:
        pc = ksp.getPC()
        pc.setType(preconditioner)
        if preconditioner == "ilu":
            # https://en.wikipedia.org/wiki/Incomplete_LU_factorization#Generalizations
            # level 1 - 3x slower on pancreas (200ms -> 600ms)
            # however, for badly conditioned problems, it works much better
            pc.setFactorLevels(1)
        elif preconditioner == "lu":
            ksp.setType(PETSc.KSP.Type.PREONLY)

    ksp.setOperators(A)
    ksp.setFromOptions()

    x, b = A.getVecs()

    return ksp, x, b


@functools.singledispatch
def _solve_many_sparse_problems_petsc(
    mat_b: sp.csc_matrix,
    _mat_a: Union[np.ndarray, sp.spmatrix],
    _solver: Optional[str] = None,
    _preconditioner: Optional[str] = None,
    _tol: float = 1e-5,
    _queue: Optional[Queue] = None,
) -> Tuple[np.ndarray, int]:
    raise NotImplementedError(f"Not implemented for type `{type(mat_b).__name__!r}`.")


@_solve_many_sparse_problems_petsc.register(np.ndarray)
def _(
    mat_b: np.ndarray,
    mat_a: Union[np.ndarray, sp.spmatrix],
    solver: Optional[str] = None,
    preconditioner: Optional[str] = None,
    tol: float = 1e-5,
    _queue: Optional[Queue] = None,
) -> Tuple[np.ndarray, int]:
    if mat_b.ndim not in (1, 2) or (mat_b.ndim == 2 and mat_b.shape[1] != 1):
        raise ValueError(f"Expected either a vector or a matrix with `1` column, got `{mat_b.shape}`.")

    if solver == "direct":  # this can sometimes happen
        solver = None

    ksp, x, b = _create_solver(mat_a, solver, preconditioner=preconditioner, tol=tol)
    b.setArray(mat_b.squeeze())

    ksp.solve(b, x)
    # `is_converged` in PETSc >= 3.20
    converged = ksp.is_converged if hasattr(ksp, "is_converged") else ksp.converged

    return np.atleast_1d(x.getArray().copy().squeeze()), int(converged)


@_solve_many_sparse_problems_petsc.register(sp.csc_matrix)
def _(
    mat_b: sp.csc_matrix,
    mat_a: Union[np.ndarray, sp.spmatrix],
    solver: Optional[str],
    preconditioner: Optional[str],
    tol: float,
    queue: Queue,
) -> Tuple[np.ndarray, int]:
    ksp, x, b = _create_solver(mat_a, solver, preconditioner=preconditioner, tol=tol)
    xs, converged = [], 0

    for value in mat_b:
        x.set(1)  # reset the solution

        b.set(0)
        b.setValues(value.indices, value.data)

        ksp.solve(b, x)

        xs.append(np.atleast_1d(x.getArray().copy().squeeze()))
        # `is_converged` in PETSc >= 3.20
        converged += ksp.is_converged if hasattr(ksp, "is_converged") else ksp.converged

        if queue is not None:
            queue.put(1)

    if queue is not None:
        queue.put(None)

    return np.stack(xs, axis=1), converged


def _solve_many_sparse_problems(
    mat_b: sp.spmatrix,
    mat_a: sp.spmatrix,
    solver: LinSolver,
    tol: float,
    queue: Queue,
) -> Tuple[np.ndarray, int]:
    """Solve ``mat_a * x = mat_b`` efficiently using an iterative solver.

    This is a utility function which is optimized for the case of ``mat_a`` and ``mat_b`` being sparse,
    and columns in ``mat_b`` being related. In that case, we can treat each column of ``mat_b`` as a
    separate linear problem and solve that efficiently using iterative solvers that exploit sparsity.

    Parameters
    ----------
    mat_b
        Matrix of shape `n x m`, with m << n.
    mat_a
        Matrix of shape `n x n`. We make no assumptions on `mat_a` being symmetric or positive definite.
    solver
        Solver to use for the linear problem. Valid options can be found in :func:`scipy.sparse.linalg`.
    tol
        The relative convergence tolerance, relative decrease in the (possibly preconditioned) residual norm .
    queue
        Queue used to signal when a solution has been computed.

    Returns
    -------
    Matrix of shape `n x m`. Each column in the resulting matrix corresponds to the solution
    of one of the sub-problems defined via columns in ``mat_b``.

    Number of converged solutions.
    """
    # initialise solution list and info list
    x_list, n_converged = [], 0
    kwargs = {} if solver is not sp.linalg.gmres else {"atol": "legacy"}  # get rid of the warning

    for b in mat_b:
        # actually call the solver for the current sub-problem
        x, info = solver(mat_a, b.toarray().flatten(), tol=tol, x0=None, **kwargs)

        # append solution and info
        x_list.append(np.atleast_1d(x))
        n_converged += info == 0

        if queue is not None:
            queue.put(1)

    if queue is not None:
        queue.put(None)

    return np.stack(x_list, axis=1), n_converged


def _petsc_direct_solve(
    mat_a: Union[np.ndarray, sp.spmatrix],
    mat_b: Optional[Union[sp.spmatrix, np.ndarray]] = None,
    tol: float = 1e-5,
    **kwargs,
) -> np.ndarray:
    from petsc4py import PETSc

    if mat_b is None:
        if mat_a.shape[0] != mat_a.shape[1]:
            raise ValueError("Matrix `A` is not square, unable to invert.")
        # matrix inversion
        B = PETSc.Mat().create()
        B.setSizes(mat_a.shape)
        B.setType(PETSc.Mat.Type.DENSE)
        B.setFromOptions()
        B.setUp()

        start, end = B.getOwnershipRange()
        for i in range(start, end):
            B[i, i] = 1
        B.assemble()
    else:
        if mat_b.ndim == 1 or (mat_b.ndim == 2 and mat_b.shape[1] == 1):
            if sp.issparse(mat_b):
                mat_b = mat_b.toarray()

            res, converged = _solve_many_sparse_problems_petsc(mat_b, mat_a=mat_a, tol=tol, **kwargs)
            if not converged:
                logg.warning(
                    f"The solution for system "
                    f"`A{list(mat_a.shape)} * x[{mat_b.shape[0]}] = b[{mat_b.shape[0]}]` "
                    f"did not converge"
                )

            return res[:, None]

        B = _create_petsc_matrix(mat_b, as_dense=True)

    A = _create_petsc_matrix(mat_a)

    x = PETSc.Mat().create()
    x.setSizes((A.getSize()[1], B.getSize()[1]))
    x.setType(PETSc.Mat.Type.DENSE)
    x.setFromOptions()
    x.setUp()
    x.assemble()

    ksp = PETSc.KSP().create()
    ksp.setType(PETSc.KSP.Type.PREONLY)
    ksp.setTolerances(rtol=tol)
    ksp.setOperators(A, A)

    pc = ksp.getPC()
    pc.setType(PETSc.PC.Type.LU)
    pc.setFromOptions()
    pc.setUp()

    ksp.setFromOptions()
    ksp.setUp()

    factored_matrix = pc.getFactorMatrix()

    # solve
    factored_matrix.matSolve(B, x)

    res = np.array(x.getDenseArray(), copy=True)
    # `is_converged` in PETSc >= 3.20
    converged = ksp.is_converged if hasattr(ksp, "is_converged") else ksp.converged

    if not converged:
        logg.debug(
            f"The solution for system "
            f"`A{list(A.getSize())} * X{list(x.getSize())} = B{list(B.getSize())}` "
            f"did not converge"
        )

    return res


def _solve_lin_system(
    mat_a: Union[np.ndarray, sp.spmatrix],
    mat_b: Union[np.ndarray, sp.spmatrix],
    solver: str = _DEFAULT_SOLVER,
    use_petsc: bool = False,
    preconditioner: Optional[str] = None,
    n_jobs: Optional[int] = None,
    backend: str = DEFAULT_BACKEND,
    tol: float = 1e-5,
    use_eye: bool = False,
    show_progress_bar: bool = True,
) -> np.ndarray:
    """Solve ``mat_a * x = mat_b`` efficiently using either iterative or direct methods.

    This is a utility function which is optimized for the case of ``mat_a`` and ``mat_b`` being sparse,
    and columns in ``mat_b`` being related. In that case, we can treat each column of ``mat_b`` as a
    separate linear problem and solve that efficiently using iterative solvers that exploit sparsity.

    If the columns of ``mat_b`` are related, we can use the solution of the previous problem as an
    initial guess for the next problem. Further, we parallelize the individual problems for each
    column in ``mat_b`` and solve them on separate kernels.

    In case ``mat_a`` is either not sparse, or very small, or ``mat_b`` has very many columns, it makes
    sense to use a direct solver instead which computes a matrix factorization and thereby solves all
    sub-problems at the same time.

    Parameters
    ----------
    mat_a
        Matrix of shape `n x n`. We make no assumptions on ``mat_a`` being symmetric or positive definite.
    mat_b
        Matrix of shape `n x m`, with m << n.
    solver
        Solver to use for the linear problem. Options are `'direct', 'gmres', 'lgmres', 'bicgstab' or 'gcrotmk'`
        when ``use_petsc`` or one of `petsc4py.PETSc.KPS.Type` otherwise.

        Information on the :mod:`scipy` iterative solvers can be found in :func:`scipy.sparse.linalg` or
        for the :mod:`petsc4py` solver in https://www.mcs.anl.gov/petsc/documentation/linearsolvertable.html.
    use_petsc
        Whether to use solvers from :mod:`petsc4py` instead of :mod:`scipy`. Recommended for large problems.
    preconditioner
        Preconditioner to use when ``use_petsc=True``. For available preconditioners, see `petsc4py.PETSc.PC.Type`.
    n_jobs
        Number of parallel jobs to use when ``use_petsc=True``. For small, quickly-solvable problems,
        we recommend high number (>=8) of cores in order to fully saturate them.
    backend
        Which backend to use for multiprocessing. See :class:`joblib.Parallel` for valid options.
    tol
        The relative convergence tolerance, relative decrease in the (possibly preconditioned) residual norm .
    use_eye
        Solve ``(I - mat_a) * x = mat_b`` instead.
    show_progress_bar
        Whether to show progress bar when the solver isn't a direct one.

    Returns
    -------
    Matrix of shape `n x m`. Each column corresponds to the solution of one of the sub-problems
    defined via columns in ``mat_b``.
    """

    def extractor(res_converged: List[Tuple[np.ndarray, int]]) -> Tuple[np.ndarray, int]:
        res, converged = zip(*res_converged)
        return np.hstack(res), sum(converged)

    n_jobs = _get_n_cores(n_jobs, n_jobs=None)

    if use_petsc and not _is_petsc_slepc_available():
        global _PETSC_ERROR_MSG_SHOWN
        if not _PETSC_ERROR_MSG_SHOWN:
            _PETSC_ERROR_MSG_SHOWN = True
            logg.warning(_PETSC_ERROR_MSG.format(_DEFAULT_SOLVER))
        solver = _DEFAULT_SOLVER
        use_petsc = False

    if use_eye:
        mat_a = (sp.eye(mat_a.shape[0]) if sp.issparse(mat_a) else np.eye(mat_a.shape[0])) - mat_a

    if solver == "direct":
        if use_petsc:
            logg.debug("Solving the linear system directly using `PETSc`")
            return _petsc_direct_solve(mat_a, mat_b, solver=solver, preconditioner=preconditioner, tol=tol)

        if sp.issparse(mat_a):
            logg.debug("Densifying `A` for `scipy` direct solver")
            mat_a = mat_a.toarray()
        if sp.issparse(mat_b):
            logg.debug("Densifying `B` for `scipy` direct solver")
            mat_b = mat_b.toarray()

        logg.debug("Solving the linear system directly using `scipy`")

        return solve(mat_a, mat_b)

    if use_petsc:
        if not sp.isspmatrix_csr(mat_a):
            mat_a = sp.csr_matrix(mat_a)

        mat_b = mat_b.T
        if not sp.isspmatrix_csc(mat_b):
            mat_b = sp.csc_matrix(mat_b)

        # as_array causes an issue, because it's called like this np.array([(NxM), (NxK), ....]
        # in the end, we want array of shape Nx(M + K + ...) - this is ensured by the extractor
        logg.debug(
            f"Solving the linear system using `PETSc` solver `{('gmres' if solver is None else solver)!r}` "
            f"on `{n_jobs}` core(s) with {'no' if preconditioner is None else preconditioner} preconditioner and "
            f"`tol={tol}`"
        )

        mat_x, n_converged = parallelize(
            _solve_many_sparse_problems_petsc,
            mat_b,
            n_jobs=n_jobs,
            backend=backend,
            as_array=False,
            extractor=extractor,
            show_progress_bar=show_progress_bar,
        )(mat_a, solver=solver, preconditioner=preconditioner, tol=tol)
    elif solver in _AVAIL_ITER_SOLVERS:
        if not sp.issparse(mat_a):
            logg.debug("Sparsifying `A` for iterative solver")
            mat_a = sp.csr_matrix(mat_a)

        mat_b = mat_b.T
        if not sp.issparse(mat_b):
            logg.debug("Sparsifying `B` for iterative solver")
            mat_b = sp.csr_matrix(mat_b)

        logg.debug(
            f"Solving the linear system using `scipy` solver `{solver!r}` on `{n_jobs} cores(s)` with `tol={tol}`"
        )

        mat_x, n_converged = parallelize(
            _solve_many_sparse_problems,
            mat_b,
            n_jobs=n_jobs,
            backend=backend,
            as_array=False,
            extractor=extractor,
            show_progress_bar=show_progress_bar,
        )(mat_a, solver=_AVAIL_ITER_SOLVERS[solver], tol=tol)

    else:
        raise ValueError(f"Invalid solver `{solver!r}`.")

    if n_converged != mat_b.shape[0]:
        logg.warning(f"`{mat_b.shape[0] - n_converged}` solution(s) did not converge")

    return mat_x


def _is_petsc_slepc_available() -> bool:
    try:
        import petsc4py  # noqa
        import slepc4py  # noqa

        return True
    except ImportError:
        return False



import collections
import datetime
import enum
import pathlib
import types
from pathlib import Path
from typing import Any, Dict, List, Literal, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import scipy.sparse as sp
from pandas.api.types import infer_dtype

import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import ListedColormap, Normalize
from matplotlib.ticker import StrMethodFormatter

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._colors import _create_categorical_colors, _get_black_or_white
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import ModeEnum
from cellrank._utils._key import Key
from cellrank._utils._lineage import Lineage
from cellrank._utils._utils import (
    _eigengap,
    _fuzzy_to_discrete,
    _series_from_one_hot_matrix,
    save_fig,
)
from cellrank.estimators.mixins import EigenMixin, LinDriversMixin, SchurMixin
from cellrank.estimators.mixins._utils import SafeGetter, StatesHolder, logger, shadow
from cellrank.estimators.terminal_states._term_states_estimator import (
    TermStatesEstimator,
)
from cellrank.kernels._base_kernel import KernelExpression

__all__ = ["GPCCA"]


class TermStatesMethod(ModeEnum):
    EIGENGAP = enum.auto()
    EIGENGAP_COARSE = enum.auto()
    TOP_N = enum.auto()
    STABILITY = enum.auto()


class CoarseTOrder(ModeEnum):
    STABILITY = enum.auto()  # diagonal
    INCOMING = enum.auto()
    OUTGOING = enum.auto()
    STAT_DIST = enum.auto()


@d.dedent
class GPCCA(TermStatesEstimator, LinDriversMixin, SchurMixin, EigenMixin):
    """Generalized Perron Cluster Cluster Analysis (GPCCA) :cite:`reuter:18,reuter:19`.

    .. seealso::
        - See :doc:`../../../notebooks/tutorials/estimators/600_initial_terminal` on how to compute the
          :attr:`initial <initial_states>` and :attr:`terminal <terminal_states>` states.
        - See :doc:`../../../notebooks/tutorials/estimators/700_fate_probabilities` on how to compute the
          :attr:`fate_probabilities` and :attr:`lineage_drivers`.

    This is our main and recommended estimator implemented in `pyGPCCA <https://pygpcca.readthedocs.io/en/latest/>`_ .
    Use it to compute macrostates, automatically and semi-automatically classify these as initial, intermediate and
    terminal states, compute fate probabilities towards macrostates, uncover driver genes, and much more. To compute and
    classify macrostates, we run the GPCCA algorithm under the hood, which returns a soft assignment of cells
    to macrostates, as well as a coarse-grained transition matrix among the set of macrostates
    :cite:`reuter:18,reuter:19`. This estimator allows you to inject prior knowledge where available
    to guide the identification of initial, intermediate and terminal states.

    Parameters
    ----------
    %(base_estimator.parameters)s
    """

    def __init__(
        self,
        object: Union[str, bool, np.ndarray, sp.spmatrix, AnnData, KernelExpression],
        **kwargs: Any,
    ):
        super().__init__(object=object, **kwargs)
        self._macrostates = StatesHolder()
        self._coarse_init_dist: Optional[pd.Series] = None
        self._coarse_stat_dist: Optional[pd.Series] = None
        self._coarse_tmat: Optional[pd.DataFrame] = None
        self._tsi: Optional[AnnData] = None

    @property
    @d.get_summary(base="gpcca_macro")
    def macrostates(self) -> Optional[pd.Series]:
        """Macrostates of the transition matrix."""
        return self._macrostates.assignment

    @property
    @d.get_summary(base="gpcca_macro_memberships")
    def macrostates_memberships(self) -> Optional[Lineage]:
        """Macrostate memberships.

        Soft assignment of microstates (cells) to macrostates.
        """
        return self._macrostates.memberships

    @property
    @d.get_summary(base="gpcca_init_states_memberships")
    def initial_states_memberships(self) -> Optional[Lineage]:
        """Initial states memberships.

        Soft assignment of cells to initial states.
        """
        return self._init_states.memberships

    @property
    @d.get_summary(base="gpcca_term_states_memberships")
    def terminal_states_memberships(self) -> Optional[Lineage]:
        """Terminal states memberships.

        Soft assignment of cells to terminal states.
        """
        return self._term_states.memberships

    @property
    @d.get_summary(base="gpcca_coarse_init")
    def coarse_initial_distribution(self) -> Optional[pd.Series]:
        """Coarse-grained initial distribution."""
        return self._coarse_init_dist

    @property
    @d.get_summary(base="gpcca_coarse_stat")
    def coarse_stationary_distribution(self) -> Optional[pd.Series]:
        """Coarse-grained stationary distribution."""
        return self._coarse_stat_dist

    @property
    @d.get_summary(base="gpcca_coarse_tmat")
    def coarse_T(self) -> Optional[pd.DataFrame]:
        """Coarse-grained transition matrix."""
        return self._coarse_tmat

    @d.get_sections(base="gpcca_compute_macro", sections=["Parameters", "Returns"])
    @d.dedent
    def compute_macrostates(
        self,
        n_states: Optional[Union[int, Sequence[int]]] = None,
        n_cells: Optional[int] = 30,
        cluster_key: Optional[str] = None,
        **kwargs: Any,
    ) -> "GPCCA":
        """Compute the macrostates.

        Parameters
        ----------
        n_states
            Number of macrostates to compute. If a :class:`~typing.Sequence`, use the *minChi*
            criterion :cite:`reuter:18`. If :obj:`None`, use the `eigengap <https://en.wikipedia.org/wiki/Eigengap>`__
            heuristic.
        %(n_cells)s
        cluster_key
            If a key to cluster labels is given, names and colors of the states will be associated with the clusters.
        kwargs
            Keyword arguments for :meth:`compute_schur`.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`macrostates` - %(gpcca_macro.summary)s
        - :attr:`macrostates_memberships` - %(gpcca_macro_memberships.summary)s
        - :attr:`coarse_T` - %(gpcca_coarse_tmat.summary)s
        - :attr:`coarse_initial_distribution` - %(gpcca_coarse_init.summary)s
        - :attr:`coarse_stationary_distribution` - %(gpcca_coarse_stat.summary)s
        - :attr:`schur_vectors` - %(schur_vectors.summary)s
        - :attr:`schur_matrix` - %(schur_matrix.summary)s
        - :attr:`eigendecomposition` - %(eigen.summary)s
        """
        n_states = self._n_states(n_states)
        if n_states == 1:
            self._compute_one_macrostate(
                n_cells=n_cells,
                cluster_key=cluster_key,
            )
            return self

        if self._gpcca is None or kwargs:
            self.compute_schur(n_states, **kwargs)
        n_states = self._validate_n_states(n_states)

        if self._gpcca._p_X.shape[1] < n_states:
            # precomputed X
            logg.warning(
                f"Requested more macrostates `{n_states}` than available "
                f"Schur vectors `{self._gpcca._p_X.shape[1]}`. Recomputing the decomposition"
            )

        start = logg.info(f"Computing `{n_states}` macrostates")
        try:
            self._gpcca = self._gpcca.optimize(m=n_states)
        except ValueError as e:
            if "will split complex conjugate eigenvalues" not in str(e):
                raise
            # this is the following case - we have 4 Schur vectors, user requests 5 states, but it splits the conj. ev.
            # in the try block, Schur decomposition with 5 vectors is computed, but it fails (no way of knowing)
            # so in this case, we increase it by 1
            logg.warning(
                f"Unable to compute macrostates with `n_states={n_states}` because it will "
                f"split complex conjugate eigenvalues. Using `n_states={n_states + 1}`"
            )
            self._gpcca = self._gpcca.optimize(m=n_states + 1)

        self._set_macrostates(
            memberships=self._gpcca.memberships,
            n_cells=n_cells,
            cluster_key=cluster_key,
            params=self._create_params(),
            time=start,
        )
        return self

    def predict(self, *args: Any, **kwargs: Any) -> "GPCCA":
        """Alias for :meth:`predict_terminal_states`.

        Parameters
        ----------
        args
            Positional arguments.
        kwargs
            Keyword arguments.

        Returns
        -------
        Same as :meth:`predict_terminal_states`.
        """
        return self.predict_terminal_states(*args, **kwargs)

    @d.dedent
    def predict_terminal_states(
        self,
        method: Literal["stability", "top_n", "eigengap", "eigengap_coarse"] = TermStatesMethod.STABILITY,
        n_cells: int = 30,
        alpha: Optional[float] = 1,
        stability_threshold: float = 0.96,
        n_states: Optional[int] = None,
        allow_overlap: bool = False,
    ) -> "GPCCA":
        """Automatically select terminal states from macrostates.

        Parameters
        ----------
        method
            How to select the terminal states. Valid option are:

            - ``'eigengap'`` - select the number of states based on the
              `eigengap <https://en.wikipedia.org/wiki/Eigengap>`__ of :attr:`transition_matrix`.
            - ``'eigengap_coarse'`` - select the number of states based on the *eigengap* of the diagonal
              of :attr:`coarse_T`.
            - ``'top_n'`` - select top ``n_states`` based on the probability of the diagonal of :attr:`coarse_T`.
            - ``'stability'`` - select states which have a stability >= ``stability_threshold``.
              The stability is given by the diagonal elements of :attr:`coarse_T`.
        %(n_cells)s
        alpha
            Weight given to the deviation of an eigenvalue from one.
            Only used when ``method = 'eigengap'`` or ``method = 'eigengap_coarse'``.
        stability_threshold
            Threshold used when ``method = 'stability'``.
        n_states
            Number of states used when ``method = 'top_n'``.
        %(allow_overlap)s

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`terminal_states` - %(tse_term_states.summary)s
        - :attr:`terminal_states_probabilities` - %(tse_term_states_probs.summary)s
        - :attr:`terminal_states_memberships` - %(gpcca_term_states_memberships.summary)s
        """
        if self.macrostates is None:
            raise RuntimeError("Compute macrostates first as `.compute_macrostates()`.")

        if len(self.macrostates.cat.categories) == 1:
            logg.warning("Found only one macrostate, making it the singular terminal state")
            return self.set_terminal_states(
                states=None,
                n_cells=n_cells,
                allow_overlap=allow_overlap,
                params=self._create_params(),
            )

        method = TermStatesMethod(method)
        eig = self.eigendecomposition
        coarse_T = self.coarse_T

        # fmt: off
        if method == TermStatesMethod.EIGENGAP:
            if eig is None:
                raise RuntimeError("Compute eigendecomposition first as `.compute_eigendecomposition()`.")
            n_states = _eigengap(eig["D"], alpha=alpha) + 1
        elif method == TermStatesMethod.EIGENGAP_COARSE:
            if coarse_T is None:
                raise RuntimeError("Compute macrostates first as `.compute_macrostates()`.")
            n_states = _eigengap(np.sort(np.diag(coarse_T)[::-1]), alpha=alpha)
        elif method == TermStatesMethod.TOP_N:
            if n_states is None:
                raise ValueError("Expected `n_states != None` for `method='top_n'`.")
            if n_states <= 0:
                raise ValueError(f"Expected `n_states` to be positive, found `{n_states}`.")
        elif method == TermStatesMethod.STABILITY:
            if stability_threshold is None:
                raise ValueError("Expected `stability_threshold != None` for `method='stability'`.")
            stability = pd.Series(np.diag(coarse_T), index=coarse_T.columns)
            names = stability[stability.values >= stability_threshold].index
            return self.set_terminal_states(
                names,
                n_cells=n_cells,
                allow_overlap=allow_overlap,
                params=self._create_params()
            )
        else:
            raise NotImplementedError(f"Method `{method}` is not yet implemented.")
        # fmt: on

        names = coarse_T.columns[np.argsort(np.diag(coarse_T))][-n_states:]
        return self.set_terminal_states(
            names,
            n_cells=n_cells,
            allow_overlap=allow_overlap,
            params=self._create_params(),
        )

    @d.dedent
    def predict_initial_states(self, n_states: int = 1, n_cells: int = 30, allow_overlap: bool = False) -> "GPCCA":
        """Compute initial states from macrostates using :attr:`coarse_stationary_distribution`.

        Parameters
        ----------
        n_states
            Number of initial states.
        %(n_cells)s
        %(allow_overlap)s

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`initial_states` - %(tse_init_states.summary)s
        - :attr:`initial_states_probabilities` - %(tse_init_states_probs.summary)s
        - :attr:`initial_states_memberships` - %(gpcca_init_states_memberships.summary)s
        """
        if n_states <= 0:
            raise ValueError(f"Expected `n_states` to be positive, found `{n_states}`.")
        if n_cells <= 0:
            raise ValueError(f"Expected `n_cells` to be positive, found `{n_cells}`.")

        probs = self.macrostates_memberships
        if probs is None:
            raise RuntimeError("Compute macrostates first as `.compute_macrostates()`.")
        if n_states > probs.shape[1]:
            raise ValueError(
                f"Requested `{n_states}` initial states, but only `{probs.shape[1]}` macrostates have been computed."
            )

        if probs.shape[1] == 1:
            return self.set_initial_states(states=None, n_cells=n_cells, allow_overlap=allow_overlap)

        stat_dist = self.coarse_stationary_distribution
        if stat_dist is None:
            raise RuntimeError("No coarse-grained stationary distribution found.")

        states = list(stat_dist.iloc[np.argsort(stat_dist)][:n_states].index)
        return self.set_initial_states(states, n_cells=n_cells, allow_overlap=allow_overlap)

    @d.dedent
    def set_terminal_states(
        self,
        states: Optional[Union[str, Sequence[str], Dict[str, Sequence[str]], pd.Series]] = None,
        n_cells: int = 30,
        allow_overlap: bool = False,
        cluster_key: Optional[str] = None,
        **kwargs: Any,
    ) -> "GPCCA":
        """Set the :attr:`terminal_states`.

        Parameters
        ----------
        states
            Which states to select. Valid options are:

            - :class:`str`, :class:`~typing.Sequence` - subset of :attr:`macrostates`. Multiple states can be
              combined using ``','``, such as ``['Alpha, Beta', 'Epsilon']``.
            - :class:`dict` - keys correspond to terminal states and values to cell IDs in
              :attr:`~anndata.AnnData.obs_names`.
            - :class:`~pandas.Series` - categorical series where each category corresponds to a macrostate.
              `NaN` values mark cells that should not be marked as :attr:`terminal_states`.
            - :obj:`None` - select all :attr:`macrostates`.
        %(n_cells)s
        %(allow_overlap)s
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` to associate names and colors with :attr:`terminal_states`.
            Each state will be given the name and color corresponding to the cluster it mostly overlaps with.
            Only used when ``states`` is a :class:`dict` or :class:`~pandas.Series`.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`terminal_states` - %(tse_term_states.summary)s
        - :attr:`terminal_states_probabilities` - %(tse_term_states_probs.summary)s
        - :attr:`terminal_states_memberships` - %(gpcca_term_states_memberships.summary)s
        """
        if isinstance(states, (dict, pd.Series)):
            return super().set_terminal_states(states, cluster_key=cluster_key, allow_overlap=allow_overlap, **kwargs)

        if n_cells <= 0:
            raise ValueError(f"Expected `n_cells` to be positive, found `{n_cells}`.")

        memberships = self.macrostates_memberships
        if memberships is None:
            raise RuntimeError("Compute macrostates first as `.compute_macrostates()`.")

        if states is None:
            states = memberships.names
        elif isinstance(states, str):
            states = [states]
        if not len(states):  # unset the states
            raise ValueError("No macrostates have been selected.")

        is_singleton = memberships.shape[1] == 1
        memberships = memberships[list(states)].copy()

        states = self._create_states(memberships, n_cells=n_cells, check_row_sums=False)
        if is_singleton:
            colors = self._macrostates.colors.copy()
            probs = memberships.X.squeeze() / memberships.X.max()
        else:
            colors = memberships[list(states.cat.categories)].colors
            probs = (memberships.X / memberships.X.max(axis=0)).max(axis=1)
        probs = pd.Series(probs, index=self.adata.obs_names)

        self._write_states(
            "terminal",
            states=states,
            colors=colors,
            probs=probs,
            memberships=memberships,
            params=kwargs.pop("params", {}),
            allow_overlap=allow_overlap,
            **kwargs,
        )
        return self

    @d.dedent
    def set_initial_states(
        self,
        states: Optional[Union[str, Sequence[str], Dict[str, Sequence[str]], pd.Series]] = None,
        n_cells: int = 30,
        allow_overlap: bool = False,
        cluster_key: Optional[str] = None,
        **kwargs: Any,
    ) -> "GPCCA":
        """Set the :attr:`initial_states`.

        Parameters
        ----------
        states
            Which states to select. Valid options are:

            - :class:`str`, :class:`~typing.Sequence` - subset of :attr:`macrostates`. Multiple states can be
              combined using ``','``, such as ``['Alpha, Beta', 'Epsilon']``.
            - :class:`dict` - keys correspond to initial states and values to cell IDs in
              :attr:`~anndata.AnnData.obs_names`.
            - :class:`~pandas.Series` - categorical series where each category corresponds to a macrostate.
              `NaN` values mark cells that should not be marked as :attr:`initial_states`.
            - :obj:`None` - select all :attr:`macrostates`.
        %(n_cells)s
        %(allow_overlap)s
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` to associate names and colors with :attr:`initial_states`.
            Each state will be given the name and color corresponding to the cluster it mostly overlaps with.
            Only used when ``states`` is a :class:`dict` or :class:`~pandas.Series`.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`initial_states` - %(tse_init_states.summary)s
        - :attr:`initial_states_probabilities` - %(tse_init_states_probs.summary)s
        - :attr:`initial_states_memberships` - %(gpcca_init_states_memberships.summary)s
        """
        if isinstance(states, (pd.Series, dict)):
            return super().set_initial_states(states, cluster_key=cluster_key, allow_overlap=allow_overlap, **kwargs)

        if n_cells <= 0:
            raise ValueError(f"Expected `n_cells` to be positive, found `{n_cells}`.")

        memberships = self.macrostates_memberships
        if memberships is None:
            raise RuntimeError("Compute macrostates first as `.compute_macrostates()`.")

        if states is None:
            states = memberships.names
        elif isinstance(states, str):
            states = [states]
        if not len(states):  # unset the states
            raise ValueError("No macrostates have been selected.")

        is_singleton = memberships.shape[1] == 1
        memberships = memberships[list(states)].copy()

        states = self._create_states(memberships, n_cells=n_cells, check_row_sums=False)
        if is_singleton:
            colors = self._macrostates.colors.copy()
            probs = memberships.X.squeeze() / memberships.X.max()
        else:
            colors = memberships[list(states.cat.categories)].colors
            probs = (memberships.X / memberships.X.max(axis=0)).max(axis=1)
        probs = pd.Series(probs, index=self.adata.obs_names)

        self._write_states(
            "initial",
            states=states,
            colors=colors,
            probs=probs,
            memberships=memberships,
            params=kwargs.pop("params", {}),
            allow_overlap=allow_overlap,
            **kwargs,
        )
        return self

    # TODO: Add definition/link to paper.
    def tsi(
        self,
        n_macrostates: int,
        terminal_states: Optional[List[str]] = None,
        cluster_key: Optional[str] = None,
        **kwargs: Any,
    ) -> float:
        """Compute terminal state identification (TSI) score.

        Parameters
        ----------
        n_macrostates
            Maximum number of macrostates to consider.
        terminal_states
            List of terminal states.
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` defining cluster labels including terminal states.
        kwargs
            Keyword arguments passed to :meth:`compute_macrostates` function.

        Returns
        -------
        Returns TSI score.
        """
        tsi_precomputed = (self._tsi is not None) and (self._tsi[:, "number_of_macrostates"].X.max() >= n_macrostates)
        if terminal_states is not None:
            tsi_precomputed = tsi_precomputed and (set(self._tsi.uns["terminal_states"]) == set(terminal_states))
        if cluster_key is not None:
            tsi_precomputed = tsi_precomputed and (self._tsi.uns["cluster_key"] == cluster_key)

        if not tsi_precomputed:
            if terminal_states is None:
                raise RuntimeError("`terminal_states` needs to be specified to compute TSI.")
            if cluster_key is None:
                raise RuntimeError("`cluster_key` needs to be specified to compute TSI.")

            # create a new GPCCA object to avoid unsetting attributes
            # that depend on the macrostates, e.g. the terminal states
            g = self.copy(deep=True)
            macrostates = {}
            for n_states in range(n_macrostates, 0, -1):
                g = g.compute_macrostates(n_states=n_states, cluster_key=cluster_key, **kwargs)
                macrostates[n_states] = g.macrostates.cat.categories

            max_terminal_states = len(terminal_states)

            tsi_df = collections.defaultdict(list)
            for n_states, states in macrostates.items():
                n_terminal_states = (
                    states.str.replace(r"(_).*", "", regex=True).drop_duplicates().isin(terminal_states).sum()
                )
                tsi_df["number_of_macrostates"].append(n_states)
                tsi_df["identified_terminal_states"].append(n_terminal_states)

                tsi_df["optimal_identification"].append(min(n_states, max_terminal_states))

            tsi_df = AnnData(pd.DataFrame(tsi_df), uns={"terminal_states": terminal_states, "cluster_key": cluster_key})
            self._tsi = tsi_df

        tsi_df = self._tsi.to_df()
        row_mask = tsi_df["number_of_macrostates"] <= n_macrostates
        optimal_score = tsi_df.loc[row_mask, "optimal_identification"].sum()

        return tsi_df.loc[row_mask, "identified_terminal_states"].sum() / optimal_score

    @d.dedent
    def plot_tsi(
        self,
        n_macrostates: Optional[int] = None,
        x_offset: Tuple[float, float] = (0.2, 0.2),
        y_offset: Tuple[float, float] = (0.1, 0.1),
        figsize: Tuple[float, float] = (6, 4),
        dpi: Optional[int] = None,
        save: Optional[Union[str, Path]] = None,
        **kwargs: Any,
    ) -> Axes:
        """Plot terminal state identificiation (TSI).

        Requires computing TSI with :meth:`tsi` first.

        Parameters
        ----------
        n_macrostates
            Maximum number of macrostates to consider. Defaults to using all.
        x_offset
            Offset of x-axis.
        y_offset
            Offset of y-axis.
        %(plotting)s
        kwargs
            Keyword arguments for :func:`~seaborn.lineplot`.

        Returns
        -------
        Plot TSI of the kernel and an optimal identification strategy.
        """
        if self._tsi is None:
            raise RuntimeError("Compute TSI with `tsi` first as `.tsi()`.")

        tsi_df = self._tsi.to_df()
        if n_macrostates is not None:
            tsi_df = tsi_df.loc[tsi_df["number_of_macrostates"] <= n_macrostates, :]

        optimal_identification = tsi_df[["number_of_macrostates", "optimal_identification"]]
        optimal_identification = optimal_identification.rename(
            columns={"optimal_identification": "identified_terminal_states"}
        )
        optimal_identification["method"] = "Optimal identification"
        optimal_identification["line_style"] = "--"

        df = tsi_df[["number_of_macrostates", "identified_terminal_states"]]
        df["method"] = self.kernel.__class__.__name__
        df["line_style"] = "-"

        df = pd.concat([df, optimal_identification])

        fig, ax = plt.subplots(figsize=figsize, dpi=dpi, tight_layout=True)
        sns.lineplot(
            data=df,
            x="number_of_macrostates",
            y="identified_terminal_states",
            hue="method",
            style="line_style",
            drawstyle="steps-post",
            ax=ax,
            **kwargs,
        )

        ax.set_xticks(df["number_of_macrostates"].unique().astype(int))
        # Plot is generated from large to small values on the x-axis
        for label_id, label in enumerate(ax.xaxis.get_ticklabels()[::-1]):
            if ((label_id + 1) % 5 != 0) and label_id != 0:
                label.set_visible(False)
        ax.set_yticks(df["identified_terminal_states"].unique())

        x_min = df["number_of_macrostates"].min() - x_offset[0]
        x_max = df["number_of_macrostates"].max() + x_offset[1]
        y_min = df["identified_terminal_states"].min() - y_offset[0]
        y_max = df["identified_terminal_states"].max() + y_offset[1]
        ax.set(
            xlim=[x_min, x_max],
            ylim=[y_min, y_max],
            xlabel="Number of macrostates",
            ylabel="Identified terminal states",
        )

        ax.get_legend().remove()

        n_methods = len(df["method"].unique())
        handles, labels = ax.get_legend_handles_labels()
        handles[n_methods].set_linestyle("--")
        handles = handles[: (n_methods + 1)]
        labels = labels[: (n_methods + 1)]
        labels[0] = "Method"
        fig.legend(handles=handles, labels=labels, loc="lower center", ncol=(n_methods + 1), bbox_to_anchor=(0.5, -0.1))

        if save is not None:
            save_fig(fig=fig, path=save)

        return ax

    @d.dedent
    def fit(
        self,
        n_states: Optional[Union[int, Sequence[int]]] = None,
        n_cells: Optional[int] = 30,
        cluster_key: Optional[str] = None,
        **kwargs: Any,
    ) -> "GPCCA":
        """
        Prepare self for terminal states prediction.

        Parameters
        ----------
        %(gpcca_compute_macro.parameters)s

        Returns
        -------
        %(gpcca_compute_macro.returns)s
        """
        if n_states is None:
            self.compute_eigendecomposition()
            n_states = self.eigendecomposition["eigengap"] + 1
        elif isinstance(n_states, int) and n_states == 1:
            self.compute_eigendecomposition()

        n = n_states if isinstance(n_states, int) else max(n_states)
        # call explicitly since `compute_macrostates` doesn't handle the case
        # when `minChi` is used for `n_states` and `self._gpcca` is uninitialized
        _ = self.compute_schur(n, **kwargs)
        return self.compute_macrostates(n_states=n_states, cluster_key=cluster_key, n_cells=n_cells)

    @d.dedent
    @inject_docs(o=CoarseTOrder)
    def plot_coarse_T(
        self,
        show_stationary_dist: bool = True,
        show_initial_dist: bool = False,
        order: Optional[Literal["stability", "incoming", "outgoing", "stat_dist"]] = "stability",
        cmap: Union[str, ListedColormap] = "viridis",
        xtick_rotation: float = 45,
        annotate: bool = True,
        show_cbar: bool = True,
        title: Optional[str] = None,
        figsize: Tuple[float, float] = (8, 8),
        dpi: int = 80,
        save: Optional[Union[str, pathlib.Path]] = None,
        text_kwargs: Mapping[str, Any] = types.MappingProxyType({}),
        **kwargs: Any,
    ) -> None:
        """Plot the coarse-grained transition matrix.

        Parameters
        ----------
        show_stationary_dist
            Whether to show the :attr:`coarse_stationary_distribution`, if present.
        show_initial_dist
            Whether to show the :attr:`coarse_initial_distribution`.
        order
            How to order the coarse-grained transition matrix. Valid options are:

            - ``{o.STABILITY!r}`` - order by the values on the diagonal.
            - ``{o.INCOMING!r}`` - order by the incoming mass, excluding the diagonal.
            - ``{o.OUTGOING!r}`` - order by the outgoing mass, excluding the diagonal.
            - ``{o.STAT_DIST!r}`` - order by coarse stationary distribution. If not present, use ``{o.STABILITY!r}``.
        cmap
            Colormap to use.
        xtick_rotation
            Rotation of ticks on the x-axis.
        annotate
            Whether to display the text on each cell.
        show_cbar
            Whether to show the colorbar.
        title
            Title of the figure.
        %(plotting)s
        text_kwargs
            Keyword arguments for :meth:`~matplotlib.axes.Axes.text`.
        kwargs
            Keyword arguments for :meth:`~matplotlib.axes.Axes.imshow`.

        Returns
        -------
        %(just_plots)s
        """

        def order_matrix(
            order: Optional[CoarseTOrder],
        ) -> Tuple[pd.DataFrame, Optional[pd.Series], Optional[pd.Series]]:
            coarse_T = self.coarse_T
            init_d = self.coarse_initial_distribution
            stat_d = self.coarse_stationary_distribution

            if order is None:
                return coarse_T, init_d, stat_d

            order = CoarseTOrder(order)
            if order == CoarseTOrder.STAT_DIST and stat_d is None:
                order = CoarseTOrder.STABILITY
                logg.warning(
                    f"Unable to order by `{CoarseTOrder.STAT_DIST}`, no coarse stationary distribution. "
                    f"Using `order={order}`"
                )

            if order == CoarseTOrder.INCOMING:
                values = (coarse_T.sum(0) - np.diag(coarse_T)).argsort(kind="stable")
                names = values.index[values]
            elif order == CoarseTOrder.OUTGOING:
                values = (coarse_T.sum(1) - np.diag(coarse_T)).argsort(kind="stable")
                names = values.index[values]
            elif order == CoarseTOrder.STABILITY:
                names = coarse_T.index[np.argsort(np.diag(coarse_T), kind="stable")]
            elif order == CoarseTOrder.STAT_DIST:
                names = stat_d.index[stat_d.argsort(kind="stable")]
            else:
                raise NotImplementedError(f"Order `{order}` is not yet implemented.")

            coarse_T = coarse_T.loc[names][names]
            if init_d is not None:
                init_d = init_d[names]
            if stat_d is not None:
                stat_d = stat_d[names]

            return coarse_T, init_d, stat_d

        def stylize_dist(ax: Axes, data: np.ndarray, xticks_labels: Sequence[str] = ()) -> None:
            _ = ax.imshow(data, aspect="auto", cmap=cmap, norm=norm)
            for spine in ax.spines.values():
                spine.set_visible(False)

            if xticks_labels is not None:
                ax.set_xticks(np.arange(data.shape[1]))
                ax.set_xticklabels(xticks_labels)
                plt.setp(
                    ax.get_xticklabels(),
                    rotation=xtick_rotation,
                    ha="right",
                    rotation_mode="anchor",
                )
            else:
                ax.set_xticks([])
                ax.tick_params(which="both", top=False, right=False, bottom=False, left=False)

            ax.set_yticks([])

        def annotate_heatmap(im, valfmt: str = "{x:.2f}") -> None:
            # modified from matplotlib's site

            data = im.get_array()
            kw = {"ha": "center", "va": "center"}
            kw.update(**text_kwargs)

            # Get the formatter in case a string is supplied
            if isinstance(valfmt, str):
                valfmt = StrMethodFormatter(valfmt)

            # Loop over the data and create a `Text` for each "pixel".
            # Change the text's color depending on the data.
            texts = []
            for i in range(data.shape[0]):
                for j in range(data.shape[1]):
                    kw.update(color=_get_black_or_white(im.norm(data[i, j]), cmap))
                    text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
                    texts.append(text)

        def annotate_dist_ax(ax: Axes, data: np.ndarray, valfmt: str = "{x:.2f}") -> None:
            if ax is None:
                return

            if isinstance(valfmt, str):
                valfmt = StrMethodFormatter(valfmt)

            kw = {"ha": "center", "va": "center"}
            kw.update(**text_kwargs)

            for i, val in enumerate(data):
                kw.update(color=_get_black_or_white(im.norm(val), cmap))
                ax.text(
                    i,
                    0,
                    valfmt(val, None),
                    **kw,
                )

        if self.coarse_T is None:
            raise RuntimeError(
                "Compute coarse-grained transition matrix first as `.compute_macrostates()` with `n_states > 1`."
            )

        coarse_T, coarse_init_d, coarse_stat_d = order_matrix(order)
        if show_stationary_dist and coarse_stat_d is None:
            logg.warning("Coarse stationary distribution is `None`, ignoring")
            show_stationary_dist = False
        if show_initial_dist and coarse_init_d is None:
            logg.warning("Coarse initial distribution is `None`, ignoring")
            show_initial_dist = False

        hrs, wrs = [1], [1]
        if show_stationary_dist:
            hrs += [0.05]
        if show_initial_dist:
            hrs += [0.05]
        if show_cbar:
            wrs += [0.025]

        dont_show_dist = not show_initial_dist and not show_stationary_dist

        fig = plt.figure(constrained_layout=False, figsize=figsize, dpi=dpi)
        gs = plt.GridSpec(
            1 + show_stationary_dist + show_initial_dist,
            1 + show_cbar,
            height_ratios=hrs,
            width_ratios=wrs,
            wspace=0.05,
            hspace=0.05,
        )
        if isinstance(cmap, str):
            cmap = plt.get_cmap(cmap)

        ax = fig.add_subplot(gs[0, 0])
        cax = fig.add_subplot(gs[:1, -1]) if show_cbar else None
        init_ax, stat_ax = None, None

        labels = list(coarse_T.columns)

        tmp = coarse_T
        if show_initial_dist:
            tmp = np.c_[tmp, coarse_stat_d]
        if show_initial_dist:
            tmp = np.c_[tmp, coarse_init_d]

        minn, maxx = np.nanmin(tmp), np.nanmax(tmp)
        norm = Normalize(vmin=minn, vmax=maxx)

        if show_stationary_dist:
            stat_ax = fig.add_subplot(gs[1, 0])
            stylize_dist(
                stat_ax,
                np.array(coarse_stat_d).reshape(1, -1),
                xticks_labels=labels if not show_initial_dist else None,
            )
            stat_ax.yaxis.set_label_position("right")
            stat_ax.set_ylabel("stationary dist", rotation=0, ha="left", va="center")

        if show_initial_dist:
            init_ax = fig.add_subplot(gs[show_stationary_dist + show_initial_dist, 0])
            stylize_dist(init_ax, np.array(coarse_init_d).reshape(1, -1), xticks_labels=labels)

            init_ax.yaxis.set_label_position("right")
            init_ax.set_ylabel("initial dist", rotation=0, ha="left", va="center")

        im = ax.imshow(coarse_T, aspect="auto", cmap=cmap, norm=norm, **kwargs)
        ax.set_title("coarse-grained transition matrix" if title is None else title)

        if cax is not None:
            _ = ColorbarBase(
                cax,
                cmap=cmap,
                norm=norm,
                ticks=np.linspace(minn, maxx, 10),
                format="%0.3f",
            )

        ax.set_yticks(np.arange(coarse_T.shape[0]))
        ax.set_yticklabels(labels)

        ax.tick_params(
            top=False,
            bottom=dont_show_dist,
            labeltop=False,
            labelbottom=dont_show_dist,
        )

        for spine in ax.spines.values():
            spine.set_visible(False)

        if dont_show_dist:
            ax.set_xticks(np.arange(coarse_T.shape[1]))
            ax.set_xticklabels(labels)
            plt.setp(
                ax.get_xticklabels(),
                rotation=xtick_rotation,
                ha="right",
                rotation_mode="anchor",
            )
        else:
            ax.set_xticks([])

        ax.set_yticks(np.arange(coarse_T.shape[0] + 1) - 0.5, minor=True)
        ax.tick_params(which="minor", bottom=dont_show_dist, left=False, top=False)

        if annotate:
            annotate_heatmap(im)
            if show_stationary_dist:
                annotate_dist_ax(stat_ax, coarse_stat_d.values)
            if show_initial_dist:
                annotate_dist_ax(init_ax, coarse_init_d.values)

        if save:
            save_fig(fig, save)

    @d.dedent
    def plot_macrostate_composition(
        self,
        key: str,
        width: float = 0.8,
        title: Optional[str] = None,
        labelrot: float = 45,
        legend_loc: Optional[str] = "upper right out",
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[int] = None,
        save: Optional[Union[str, pathlib.Path]] = None,
        show: bool = True,
    ) -> Optional[Axes]:
        """Plot histogram of macrostates over categorical annotations.

        Parameters
        ----------
        %(adata)s
        key
            Key from :attr:`~anndata.AnnData.obs` containing categorical annotations.
        width
            Bar width in :math:`[0, 1]`.
        title
            Title of the figure. If :obj:`None`, create one automatically.
        labelrot
            Rotation of labels on x-axis.
        legend_loc
            Position of the legend. If :obj:`None`, don't show the legend.
        %(plotting)s
        show
            If `False`, return the :class:`~matplotlib.axes.Axes` object.

        Returns
        -------
        If ``show = True``, nothing, just plots, otherwise returns the axes object.
        Optionally saves it based on ``save``.
        """
        from cellrank.pl._utils import _position_legend

        macrostates = self.macrostates
        if macrostates is None:
            raise RuntimeError("Compute macrostates first as `.compute_macrostates()`.")
        if key not in self.adata.obs:
            raise KeyError(f"Data not found in `adata.obs[{key!r}]`.")
        if not isinstance(self.adata.obs[key].dtype, pd.CategoricalDtype):
            raise TypeError(
                f"Expected `adata.obs[{key!r}]` to be `categorical`, " f"found `{infer_dtype(self.adata.obs[key])}`."
            )

        mask = ~macrostates.isnull()
        df = (
            pd.DataFrame({"macrostates": macrostates, key: self.adata.obs[key]})[mask]
            .groupby([key, "macrostates"])
            .size()
        )
        try:
            cats_colors = self.adata.uns[f"{key}_colors"]
        except KeyError:
            cats_colors = _create_categorical_colors(len(self.adata.obs[key].cat.categories))
        cat_color_mapper = dict(zip(self.adata.obs[key].cat.categories, cats_colors))
        x_indices = np.arange(len(macrostates.cat.categories))
        bottom = np.zeros_like(x_indices, dtype=float)

        width = min(1, max(0, width))
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi, tight_layout=True)
        for cat, color in cat_color_mapper.items():
            frequencies = df.loc[cat]
            # do not add to legend if category is missing
            if np.sum(frequencies) > 0:
                ax.bar(
                    x_indices,
                    frequencies,
                    width,
                    label=cat,
                    color=color,
                    bottom=bottom,
                    ec="black",
                    lw=0.5,
                )
                bottom += np.array(frequencies)

        ax.set_xticks(x_indices)
        ax.set_xticklabels(
            # assuming at least 1 category
            frequencies.index,
            rotation=labelrot,
            ha="center" if labelrot in (0, 90) else "right",
        )
        y_max = bottom.max()
        ax.set_ylim([0, y_max + 0.05 * y_max])
        ax.set_yticks(np.linspace(0, y_max, 5))
        ax.margins(0.05)

        ax.set_xlabel("macrostate")
        ax.set_ylabel("frequency")
        if title is None:
            title = f"distribution over {key}"
        ax.set_title(title)
        if legend_loc not in (None, "none"):
            _position_legend(ax, legend_loc=legend_loc)

        if save is not None:
            save_fig(fig, save)

        if not show:
            return ax

    def _n_states(self, n_states: Optional[Union[int, Sequence[int]]]) -> int:
        if n_states is None:
            if self.eigendecomposition is None:
                raise RuntimeError(
                    "Compute eigendecomposition first as `.compute_eigendecomposition()` or "
                    "supply `n_states != None`."
                )
            return self.eigendecomposition["eigengap"] + 1

        # fmt: off
        if isinstance(n_states, int):
            if n_states <= 0:
                raise ValueError(f"Expected `n_states` to be positive, found `{n_states}`.")
            return n_states

        if self._gpcca is None:
            raise RuntimeError("Compute Schur decomposition first as `.compute_schur()`.")

        if not isinstance(n_states, Sequence):
            raise TypeError(f"Expected `n_states` to be a `Sequence`, found `{type(n_states).__name__!r}`.")
        if len(n_states) != 2:
            raise ValueError(f"Expected `n_states` to be of size `2`, found `{len(n_states)}`.")

        minn, maxx = sorted(n_states)
        if minn <= 1:
            minn = 2
            logg.warning(f"Minimum value must be larger than `1`, found `{minn}`. Setting `min={minn}`")
        if minn == 2:
            minn = 3
            logg.warning(
                f"In most cases, 2 clusters will always be optimal. "
                f"If you really expect 2 clusters, use `n_states=2`. Setting `min={minn}`"
            )
        # fmt: on
        maxx = max(minn + 1, maxx)

        logg.info(f"Calculating minChi criterion in interval `[{minn}, {maxx}]`")

        return int(np.arange(minn, maxx + 1)[np.argmax(self._gpcca.minChi(minn, maxx))])

    def _create_states(
        self,
        probs: Union[np.ndarray, Lineage],
        n_cells: int,
        check_row_sums: bool = False,
        return_not_enough_cells: bool = False,
    ) -> Union[pd.Series, Tuple[pd.Series, np.ndarray]]:
        if n_cells <= 0:
            raise ValueError(f"Expected `n_cells` to be positive, found `{n_cells}`.")

        discrete, not_enough_cells = _fuzzy_to_discrete(
            a_fuzzy=probs,
            n_most_likely=n_cells,
            remove_overlap=False,
            raise_threshold=0.2,
            check_row_sums=check_row_sums,
        )

        states = _series_from_one_hot_matrix(
            membership=discrete,
            index=self.adata.obs_names,
            names=probs.names if isinstance(probs, Lineage) else None,
        )

        return (states, not_enough_cells) if return_not_enough_cells else states

    def _validate_n_states(self, n_states: int) -> int:
        if self._invalid_n_states is not None and n_states in self._invalid_n_states:
            logg.warning(
                f"Unable to compute macrostates with `n_states={n_states}` because it will "
                f"split complex conjugate eigenvalues. Using `n_states={n_states + 1}`"
            )
            n_states += 1  # cannot force recomputation of the Schur decomposition
            assert n_states not in self._invalid_n_states, "Sanity check failed."

        return n_states

    def _compute_one_macrostate(
        self,
        n_cells: Optional[int],
        cluster_key: Optional[str],
    ) -> None:
        start = logg.info("For 1 macrostate, stationary distribution is computed")

        eig = self.eigendecomposition
        if eig is not None and "stationary_dist" in eig and eig["params"]["which"] == "LR":
            stationary_dist = eig["stationary_dist"]
        else:
            self.compute_eigendecomposition(only_evals=False, which="LR")
            stationary_dist = self.eigendecomposition["stationary_dist"]

        self._set_macrostates(
            memberships=stationary_dist[:, None],
            n_cells=n_cells,
            cluster_key=cluster_key,
            check_row_sums=False,
            time=start,
        )

    @d.dedent
    def _set_macrostates(
        self,
        memberships: np.ndarray,
        n_cells: Optional[int] = 30,
        cluster_key: str = "clusters",
        check_row_sums: bool = True,
        time: Optional[datetime.datetime] = None,
        params: Dict[str, Any] = types.MappingProxyType({}),
    ) -> None:
        """Map fuzzy clustering to pre-computed annotations to get names and colors.

        Given the fuzzy clustering, we would like to select the most likely cells from each state and use these to
        give each state a name and a color by comparing with pre-computed, categorical cluster annotations.

        Parameters
        ----------
        memberships
            Fuzzy clustering.
        %(n_cells)s
        cluster_key
            Key from :attr:`~anndata.AnnData.obs` to get reference cluster annotations.
        check_row_sums
            Check whether rows in `memberships` sum to :math:`1`.
        time
            Start time of macrostates computation.
        params
            Parameters used in macrostates computation.

        Returns
        -------
        Nothing, just updates the fields as described in :meth:`compute_macrostates`.
        """
        if n_cells is None:
            # fmt: off
            logg.debug("Setting the macrostates using macrostate assignment")
            assignment = pd.Series(np.argmax(memberships, axis=1).astype(str), dtype="category")
            # sometimes, a category can be missing
            assignment = assignment.cat.reorder_categories([str(i) for i in range(memberships.shape[1])])
            # fmt: on
        else:
            logg.debug("Setting the macrostates using macrostates memberships")

            # select the most likely cells from each macrostate
            assignment, not_enough_cells = self._create_states(
                memberships,
                n_cells=n_cells,
                check_row_sums=check_row_sums,
                return_not_enough_cells=True,
            )

        # remove previous fields
        # fmt: off
        self._write_states("initial", states=None, colors=None, probs=None, memberships=None, log=False)
        self._write_states("terminal", states=None, colors=None, probs=None, memberships=None, log=False)

        assignment, colors = self._set_categorical_labels(assignment, cluster_key=cluster_key)
        memberships = Lineage(memberships, names=list(assignment.cat.categories), colors=colors)
        # fmt: on

        groups = assignment.value_counts()
        groups = groups[groups != n_cells].to_dict()
        if len(groups):
            logg.warning(
                f"The following terminal states have different number " f"of cells than requested ({n_cells}): {groups}"
            )

        self._write_macrostates(
            macrostates=assignment,
            colors=colors,
            memberships=memberships,
            time=time,
            params=params,
        )

    @logger
    @shadow
    def _write_macrostates(
        self,
        macrostates: pd.Series,
        colors: np.ndarray,
        memberships: Lineage,
        params: Dict[str, Any] = types.MappingProxyType({}),
    ) -> str:
        # fmt: off
        key = Key.obs.macrostates(self.backward)
        self._set(obj=self.adata.obs, key=key, value=macrostates)
        ckey = Key.uns.colors(key)
        self._set(obj=self.adata.uns, key=ckey, value=colors)
        mkey = Key.obsm.memberships(key)
        self._set(obj=self.adata.obsm, key=mkey, value=memberships)
        self._macrostates = self._macrostates.set(assignment=macrostates, colors=colors, memberships=memberships)
        self.params[key] = dict(params)

        names = list(macrostates.cat.categories)
        if len(names) > 1:
            # not using stationary distribution
            g = self._gpcca
            tmat = pd.DataFrame(g.coarse_grained_transition_matrix, index=names, columns=names)
            init_dist = pd.Series(g.coarse_grained_input_distribution, index=names)
            if g.coarse_grained_stationary_probability is None:
                stat_dist = None
            else:
                stat_dist = pd.Series(g.coarse_grained_stationary_probability, index=names)
            dists = pd.DataFrame({"coarse_init_dist": init_dist}, index=names)
            if stat_dist is not None:
                dists["coarse_stat_dist"] = stat_dist

            key = Key.obsm.schur_vectors(self.backward)
            self._set("_schur_vectors", obj=self.adata.obsm, key=key, value=g._p_X, shadow_only=True)
            key = Key.uns.schur_matrix(self.backward)
            self._set("_schur_matrix", obj=self.adata.uns, key=key, value=g._p_R, shadow_only=True)
            self._set("_coarse_tmat", value=tmat, shadow_only=True)
            self._set("_coarse_init_dist", value=init_dist, shadow_only=True)
            self._set("_coarse_stat_dist", value=stat_dist, shadow_only=True)
            self._set(
                obj=self.adata.uns, key=Key.uns.coarse(self.backward),
                value=AnnData(tmat, obs=dists),
            )
        else:
            for attr in ["_schur_vectors", "_schur_matrix", "_coarse_tmat", "_coarse_init_dist", "_coarse_stat_dist"]:
                self._set(attr, value=None, shadow_only=True)
            self._set(obj=self.adata.uns, key=Key.uns.coarse(self.backward), value=None)
        # fmt: on

        return (
            "Adding `.macrostates`\n"
            "       `.macrostates_memberships`\n"
            "       `.coarse_T`\n"
            "       `.coarse_initial_distribution\n"
            "       `.coarse_stationary_distribution`\n"
            "       `.schur_vectors`\n"
            "       `.schur_matrix`\n"
            "       `.eigendecomposition`\n"
            "    Finish"
        )

    @logger
    @shadow
    def _write_states(
        self,
        which: Literal["initial", "terminal"],
        states: Optional[pd.Series],
        colors: Optional[np.ndarray],
        probs: Optional[pd.Series] = None,
        memberships: Optional[Lineage] = None,
        params: Dict[str, Any] = types.MappingProxyType({}),
        allow_overlap: bool = False,
    ) -> str:
        msg = super()._write_states(
            which,
            states=states,
            colors=colors,
            probs=probs,
            params=params,
            allow_overlap=allow_overlap,
            log=False,
        )
        # fmt: off
        msg = "\n".join(msg.split("\n")[:-1])
        msg += f"\n       `.{which}_states_memberships\n    Finish`"

        self._write_fate_probabilities(None, log=False)
        # TODO(michalk8): CFLARE doesn't remove the downstream properties
        self._write_absorption_times(None, log=False)

        backward = which == "initial"
        key = Key.obsm.memberships(Key.obs.term_states(self.backward, bwd=backward))
        self._set(obj=self.adata.obsm, key=key, value=memberships)
        if backward:
            self._init_states = self._init_states.set(memberships=memberships)
        else:
            self._term_states = self._term_states.set(memberships=memberships)
        # fmt: on

        return msg

    def _read_from_adata(self, adata: AnnData, **kwargs: Any) -> bool:
        # design choice: no need to eigen/Schur-decomposition
        _ = self._read_eigendecomposition(adata, allow_missing=True)
        ok = self._read_schur_decomposition(adata, allow_missing=True)
        if not ok:
            return False

        # fmt: off
        with SafeGetter(self, allowed=KeyError) as sg:
            key = Key.obs.macrostates(self.backward)
            # TODO(michalk8): in the future, be more stringent and ensure the categories match the macro memberships
            assignment = self._get(obj=adata.obs, key=key, shadow_attr="obs", dtype=pd.Series)
            ckey = Key.uns.colors(key)
            colors = self._get(obj=adata.uns, key=ckey, shadow_attr="uns", dtype=(list, tuple, np.ndarray))
            mkey = Key.obsm.memberships(key)
            memberships = self._get(obj=adata.obsm, key=mkey, shadow_attr="obsm", dtype=(Lineage, np.ndarray))
            memberships = self._ensure_lineage_object(memberships, backward=self.backward, kind="macrostates")
            self._macrostates = StatesHolder(assignment=assignment, colors=colors, memberships=memberships)
            self.params[key] = self._read_params(key)

            tmat = adata.uns[Key.uns.coarse(self.backward)].copy()
            if not isinstance(tmat, AnnData):
                raise TypeError(
                    f"Expected coarse-grained transition matrix to be stored "
                    f"as `AnnData`, found `{type(tmat).__name__}`."
                )

            self._coarse_tmat = pd.DataFrame(tmat.X, index=tmat.obs_names, columns=tmat.obs_names)
            self._coarse_init_dist = tmat.obs["coarse_init_dist"]
            self._coarse_stat_dist = tmat.obs.get("coarse_stat_dist", None)

            self._set(obj=self._shadow_adata.uns, key=Key.uns.coarse(self.backward), value=tmat)

        if not sg.ok:
            return False

        if not super()._read_from_adata(adata, **kwargs):
            return False

        # status is based on `backward=False` by design
        for backward in [True, False]:
            with SafeGetter(self, allowed=KeyError) as sg:
                key = Key.obsm.memberships(Key.obs.term_states(self.backward, bwd=backward))
                memberships = self._get(obj=adata.obsm, key=key, shadow_attr="obsm", dtype=(np.ndarray, Lineage))
                memberships = self._ensure_lineage_object(memberships, backward=backward, kind="term_states")
                if backward:
                    self._init_states = self._init_states.set(memberships=memberships)
                else:
                    self._term_states = self._term_states.set(memberships=memberships)
                self._set(obj=self._shadow_adata.obsm, key=key, value=memberships)
        # fmt: on

        fate_prob_ok = self._read_fate_probabilities(adata)
        abs_time_ok = self._read_absorption_times(adata)
        return sg.ok and fate_prob_ok and abs_time_ok



from cellrank.estimators.mixins._fate_probabilities import FateProbsMixin
from cellrank.estimators.mixins._kernel import KernelMixin
from cellrank.estimators.mixins._lineage_drivers import LinDriversMixin
from cellrank.estimators.mixins.decomposition import EigenMixin, SchurMixin

__all__ = ["FateProbsMixin", "KernelMixin", "LinDriversMixin", "EigenMixin", "SchurMixin"]



from cellrank.models._base_model import BaseModel, FailedModel, FittedModel
from cellrank.models._gamr_model import GAMR
from cellrank.models._pygam_model import GAM
from cellrank.models._sklearn_model import SKLearnModel

__all__ = ["BaseModel", "FailedModel", "FittedModel", "GAMR", "GAM", "SKLearnModel"]



from cellrank.kernels.utils._projection import TmatProjection
from cellrank.kernels.utils._pseudotime_scheme import (
    CustomThresholdScheme,
    HardThresholdScheme,
    SoftThresholdScheme,
    ThresholdSchemeABC,
)
from cellrank.kernels.utils._random_walk import RandomWalk
from cellrank.kernels.utils._similarity import (
    Correlation,
    Cosine,
    DotProduct,
    SimilarityABC,
)
from cellrank.kernels.utils._tmat_flow import FlowPlotter
from cellrank.kernels.utils._velocity_model import Deterministic, MonteCarlo, Stochastic

__all__ = [
    "TmatProjection",
    "CustomThresholdScheme",
    "HardThresholdScheme",
    "SoftThresholdScheme",
    "RandomWalk",
    "Cosine",
    "Correlation",
    "DotProduct",
    "SimilarityABC",
    "FlowPlotter",
    "Deterministic",
    "MonteCarlo",
    "Stochastic",
]



import enum
import pathlib
import types
from typing import Any, Callable, Literal, Mapping, Optional, Sequence, Tuple, Union

import scvelo as scv

import numpy as np
import pandas as pd
from sklearn.metrics import pairwise_distances

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.colors import LinearSegmentedColormap, LogNorm

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils import Lineage
from cellrank._utils._docs import d
from cellrank._utils._enum import ModeEnum
from cellrank._utils._key import Key
from cellrank._utils._lineage import PrimingDegree
from cellrank._utils._utils import _check_collection, _unique_order_preserving, save_fig
from cellrank.pl._utils import _held_karp

__all__ = ["circular_projection"]


class LineageOrder(ModeEnum):
    DEFAULT = enum.auto()
    OPTIMAL = enum.auto()


class LabelRot(ModeEnum):
    DEFAULT = enum.auto()
    BEST = enum.auto()


Metric_T = Union[str, Callable, np.ndarray, pd.DataFrame]
_N = 200


def _get_distances(data: Union[np.ndarray, Lineage], metric: Metric_T) -> np.ndarray:
    if isinstance(data, Lineage):
        data = data.X

    if isinstance(metric, str) or callable(metric):
        metric = pairwise_distances(data.T, metric=metric)
    elif isinstance(metric, (pd.DataFrame, np.ndarray)):
        shape = (data.shape[1], data.shape[1])
        if metric.shape != shape:
            raise ValueError(
                f"Expected an `numpy.array` or `pandas.DataFrame` of shape `{shape}`, found `{metric.shape}`."
            )
    else:
        raise TypeError(
            f"Expected either metric defined by `str`, `callable` or a pairwise distance matrix of type"
            f" `numpy.ndarray` or `pandas.DataFrame`, found `{type(metric).__name__}`."
        )

    return np.asarray(metric, dtype=np.float64)


def _get_optimal_order(data: Lineage, metric: Metric_T) -> Tuple[float, np.ndarray]:
    """Solve the TSP using dynamic programming."""
    return _held_karp(_get_distances(data, metric))


@d.dedent
def circular_projection(
    adata: AnnData,
    keys: Union[str, Sequence[str]],
    backward: bool = False,
    lineages: Optional[Union[str, Sequence[str]]] = None,
    early_cells: Optional[Union[Mapping[str, Sequence[str]], Sequence[str]]] = None,
    lineage_order: Optional[Literal["default", "optimal"]] = None,
    metric: Union[str, Callable, np.ndarray, pd.DataFrame] = "correlation",
    normalize_by_mean: bool = True,
    ncols: int = 4,
    space: float = 0.25,
    use_raw: bool = False,
    text_kwargs: Mapping[str, Any] = types.MappingProxyType({}),
    label_distance: float = 1.25,
    label_rot: Union[Literal["default", "best"], float] = "best",
    show_edges: bool = True,
    key_added: Optional[str] = None,
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[int] = None,
    save: Optional[Union[str, pathlib.Path]] = None,
    **kwargs: Any,
) -> None:
    r"""Visualize fate probabilities in a circular embedding :cite:`velten:17,jaitin:14`.

    We arrange all computed terminal states evenly spaced around the unit circle and place
    cells inside the unit circle in a way that reflects their fate probabilities. In other words,
    the more certain we are that a cell will transition towards a given terminal state, the closer
    we place it to that terminal state. Uncommitted cells thus reside in the middle of the circle.

    Parameters
    ----------
    %(adata)s
    keys
        Keys in :attr:`~anndata.AnnData.obs` or :attr:`~anndata.AnnData.var_names`. Additional keys are:

        - ``'kl_divergence'`` - as in :cite:`velten:17`, computes KL-divergence between the fate probabilities
          of a cell and the average fate probabilities. See ``early_cells`` for more information.
        - ``'entropy'`` - as in :cite:`setty:19`, computes entropy over a cells fate probabilities.
    %(backward)s
    lineages
        Lineages to plot. If :obj:`None`, plot all lineages.
    early_cells
        Cell ids or a mask marking early cells used to define the average fate probabilities. If :obj:`None`,
        use all cells. Only used when ``'kl_divergence'`` is in the ``keys``. If a :class:`dict`, key specifies
        a cluster key in :attr:`~anndata.AnnData.obs` and the values specify cluster labels containing early cells.
    lineage_order
        Can be one of the following:

        - :obj:`None` - it will be determined automatically, based on the number of lineages.
        - ``'optimal'`` - order lineages optimally by solving the Traveling salesman problem (TSP).
          Recommended for <= :math:`20` lineages.
        - ``'default'`` - use the order as specified by ``lineages``.
    metric
        Metric to use when constructing pairwise distance matrix when ``lineage_order = 'optimal'``. For available
        options, see :func:`~sklearn.metrics.pairwise_distances`.
    normalize_by_mean
        If :obj:`True`, normalize each lineage by its mean probability, as done in :cite:`velten:17`.
    ncols
        Number of columns when plotting multiple ``keys``.
    space
        Horizontal and vertical space between for :func:`~matplotlib.pyplot.subplots_adjust`.
    use_raw
        Whether to access :attr:`~anndata.AnnData.raw` when there are ``keys`` in :attr:`~anndata.AnnData.var_names`.
    text_kwargs
        Keyword arguments for :meth:`~matplotlib.axes.Axes.text`.
    label_distance
        Distance at which the lineage labels will be drawn.
    label_rot
        How to rotate the labels. Valid options are:

        - ``'best'`` - rotate labels so that they are easily readable.
        - ``'default'`` - use :mod:`matplotlib`'s default.
        - :obj:`None` - same as ``'default'``.
        - :class:`float`, all labels will be rotated by this many degrees.
    show_edges
        Whether to show the edges surrounding the simplex.
    key_added
        Key in :attr:`~anndata.AnnData.obsm` where to add the circular embedding. If :obj:`None`, it will be set to
        ``'X_fate_simplex_{fwd,bwd}'``, based on the ``backward``.
    %(plotting)s
    kwargs
        Keyword arguments for :func:`~scvelo.pl.scatter`.

    Returns
    -------
    %(just_plots)s Also updates ``adata`` with the following fields:

    - :attr:`adata.obsm['{key_added}'] <anndata.AnnData.obsm>` - the circular projection.
    - :attr:`adata.obs['to_{initial,terminal}_states_{method}'] <anndata.AnnData.obs>` - the priming degree,
      if a method is present in the ``keys``.
    """
    if label_distance is not None and label_distance < 0:
        raise ValueError(f"Expected `label_distance` to be positive, found `{label_distance}`.")

    if label_rot is None:
        label_rot = LabelRot.DEFAULT
    label_rot = LabelRot(label_rot)

    suffix = "bwd" if backward else "fwd"
    if key_added is None:
        key_added = "X_fate_simplex_" + suffix

    if isinstance(keys, str):
        keys = (keys,)

    keys = _unique_order_preserving(keys)
    keys_ = _check_collection(adata, keys, "obs", key_name="Observation", raise_exc=False) + _check_collection(
        adata, keys, "var_names", key_name="Gene", raise_exc=False, use_raw=use_raw
    )
    haystack = set(PrimingDegree)
    keys = keys_ + [k for k in keys if k in haystack]
    keys = _unique_order_preserving(keys)

    if not len(keys):
        raise ValueError("No valid keys have been selected.")

    probs = Lineage.from_adata(adata, backward=backward)
    if isinstance(lineages, str):
        lineages = [lineages]
    elif lineages is None:
        lineages = probs.names
    lineages = _unique_order_preserving(lineages)
    probs = probs[lineages]
    if probs.nlin < 3:
        raise ValueError(f"Expected at least `3` lineages, found `{probs.nlin}`.")

    X = probs.X.copy()
    if normalize_by_mean:
        X /= np.mean(X, axis=0)[None, :]
        X /= X.sum(1)[:, None]
        # this happens when cells for sel. lineages sum to 1 (or when the lineage average is 0, which is unlikely)
        X = np.nan_to_num(X, nan=1.0 / probs.nlin, copy=False)

    if lineage_order is None:
        lineage_order = LineageOrder.OPTIMAL if 3 < probs.nlin <= 20 else LineageOrder.DEFAULT
        logg.debug(f"Set ordering to `{lineage_order}`")
    lineage_order = LineageOrder(lineage_order)

    if lineage_order == LineageOrder.OPTIMAL:
        logg.info(f"Solving TSP for `{probs.nlin}` states")
        _, order = _get_optimal_order(X, metric=metric)
    else:
        order = np.arange(probs.nlin)

    probs = probs[:, order]
    X = X[:, order]

    angle_vec = np.linspace(0, 2 * np.pi, probs.nlin, endpoint=False)
    angle_vec_sin = np.cos(angle_vec)
    angle_vec_cos = np.sin(angle_vec)

    x = np.sum(X * angle_vec_sin, axis=1)
    y = np.sum(X * angle_vec_cos, axis=1)
    adata.obsm[key_added] = np.c_[x, y]

    nrows = int(np.ceil(len(keys) / ncols))
    fig, ax = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        figsize=(ncols * 5, nrows * 5) if figsize is None else figsize,
        dpi=dpi,
    )

    fig.subplots_adjust(wspace=space, hspace=space)
    axes = np.ravel([ax])

    text_kwargs = dict(text_kwargs)
    text_kwargs["ha"] = "center"
    text_kwargs["va"] = "center"

    _i, lineage_key = 0, Key.obsm.fate_probs(backward)
    for _i, (k, ax) in enumerate(zip(keys, axes)):
        set_lognorm, colorbar = False, kwargs.pop("colorbar", True)
        try:
            _ = PrimingDegree(k)
            logg.debug(f"Calculating priming degree using `method={k}`")
            val = probs.priming_degree(method=k, early_cells=early_cells)
            k = f"{lineage_key}_{k}"
            adata.obs[k] = val
        except ValueError:
            # TODO(michalk8): parse the exception
            pass

        scv.pl.scatter(
            adata,
            basis=key_added,
            color=k,
            show=False,
            ax=ax,
            use_raw=use_raw,
            norm=LogNorm() if set_lognorm else None,
            colorbar=colorbar,
            **kwargs,
        )
        if colorbar and set_lognorm:
            cbar = ax.collections[0].colorbar
            cax = cbar.locator.axis
            ticks = cax.minor.locator.tick_values(cbar.vmin, cbar.vmax)
            ticks = [ticks[0], ticks[len(ticks) // 2 + 1], ticks[-1]]
            cbar.set_ticks(ticks)
            cbar.set_ticklabels([f"{t:.2f}" for t in ticks])
            cbar.update_ticks()

        patches, texts = ax.pie(
            np.ones_like(angle_vec),
            labeldistance=label_distance,
            rotatelabels=True,
            labels=probs.names[::-1],
            startangle=-360 / len(angle_vec) / 2,
            counterclock=False,
            textprops=text_kwargs,
        )

        for patch in patches:
            patch.set_visible(False)

        # clockwise
        for color, text in zip(probs.colors[::-1], texts):
            if isinstance(label_rot, (int, float, np.number)):
                text.set_rotation(label_rot)
            elif label_rot == LabelRot.BEST:
                rot = text.get_rotation()
                text.set_rotation(rot + 90 + (1 - rot // 180) * 180)
            elif label_rot != LabelRot.DEFAULT:
                raise NotImplementedError(f"Label rotation `{label_rot}` is not yet implemented.")
            text.set_color(color)

        if not show_edges:
            continue

        for i, color in enumerate(probs.colors):
            next = (i + 1) % probs.nlin
            x = 1.04 * np.linspace(angle_vec_sin[i], angle_vec_sin[next], _N)
            y = 1.04 * np.linspace(angle_vec_cos[i], angle_vec_cos[next], _N)
            points = np.array([x, y]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)

            cmap = LinearSegmentedColormap.from_list("fate_prob_cmap", [color, probs.colors[next]], N=_N)
            lc = LineCollection(segments, cmap=cmap, zorder=-1)
            lc.set_array(np.linspace(0, 1, _N))
            lc.set_linewidth(2)
            ax.add_collection(lc)

    for j in range(_i + 1, len(axes)):
        axes[j].remove()

    if save is not None:
        save_fig(fig, save)



from typing import Any, Callable, Literal, Optional, Sequence, Tuple, Union

from scvelo.preprocessing.moments import get_moments

import numpy as np
import scipy.sparse as sp

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import DEFAULT_BACKEND, Backend_t
from cellrank.kernels._base_kernel import BidirectionalKernel
from cellrank.kernels.mixins import ConnectivityMixin
from cellrank.kernels.utils import Deterministic, MonteCarlo, Stochastic
from cellrank.kernels.utils._similarity import Similarity, SimilarityABC
from cellrank.kernels.utils._velocity_model import BackwardMode, VelocityModel

__all__ = ["VelocityKernel"]


@d.dedent
class VelocityKernel(ConnectivityMixin, BidirectionalKernel):
    """Kernel which computes a transition matrix based on RNA velocity.

    .. seealso::
        - See :doc:`../../../notebooks/tutorials/kernels/200_rna_velocity` on how to
          compute the :attr:`~cellrank.kernels.VelocityKernel.transition_matrix` based on RNA velocity.

    This borrows ideas from both :cite:`manno:18` and :cite:`bergen:20`. In short, for each cell :math:`i`, we compute
    transition probabilities :math:`T_{i, j}` to each cell :math:`j` in the neighborhood of :math:`i`. We quantify
    how much the velocity vector :math:`v_i` of cell :math:`i` points towards each of its nearst neighbors. For
    this comparison, we support various schemes including cosine similarity and pearson correlation.

    Parameters
    ----------
    %(adata)s
    %(backward)s
    attr
        Attribute of :class:`~anndata.AnnData` to read from.
    xkey
        Key in :attr:`~anndata.AnnData.layers` or :attr:`~anndata.AnnData.obsm`
        where expected gene expression counts are stored.
    vkey
        Key in :attr:`~anndata.AnnData.layers` or :attr:`~anndata.AnnData.obsm` where velocities are stored.
    gene_subset
        List of genes to be used to compute transition probabilities.
        If not specified, genes from :attr:`adata.var['{vkey}_genes'] <anndata.AnnData.var>` are used.
        This feature is only available when reading from :attr:`anndata.AnnData.layers` and will be ignored otherwise.
    kwargs
        Keyword arguments for the :class:`~cellrank.kernels.Kernel`.
    """

    def __init__(
        self,
        adata: AnnData,
        backward: bool = False,
        attr: Optional[Literal["layers", "obsm"]] = "layers",
        xkey: Optional[str] = "Ms",
        vkey: Optional[str] = "velocity",
        **kwargs: Any,
    ):
        super().__init__(
            adata,
            backward=backward,
            xkey=xkey,
            vkey=vkey,
            attr=attr,
            **kwargs,
        )
        self._logits: Optional[np.ndarray] = None

    def _read_from_adata(
        self,
        xkey: Optional[str] = "Ms",
        vkey: Optional[str] = "velocity",
        attr: Optional[Literal["layers", "obsm"]] = "layers",
        gene_subset: Optional[Union[str, Sequence[str]]] = None,
        **kwargs: Any,
    ) -> None:
        super()._read_from_adata(**kwargs)

        if attr == "layers" and gene_subset is None and f"{vkey}_genes" in self.adata.var:
            gene_subset = self.adata.var[f"{vkey}_genes"]
        elif attr == "obsm" and gene_subset is not None:
            logg.warning(
                f"Found `gene_subset != None`, but it is not supported for `adata.{attr}`. Using `gene_subset = None`."
            )
            gene_subset = None

        self._xdata = self._extract_data(key=xkey, attr=attr, subset=gene_subset)
        self._vdata = self._extract_data(key=vkey, attr=attr, subset=gene_subset)
        np.testing.assert_array_equal(
            x=self._xdata.shape,
            y=self._vdata.shape,
            err_msg=f"Shape mismatch: {self._xdata.shape} vs {self._vdata.shape}",
        )

        nans = np.isnan(np.sum(self._vdata, axis=0))
        if np.any(nans):
            self._xdata = self._xdata[:, ~nans]
            self._vdata = self._vdata[:, ~nans]
        # fmt: off
        self._vexp = get_moments(self.adata, self._vdata, second_order=False).astype(np.float64, copy=False)
        self._vvar = get_moments(self.adata, self._vdata, second_order=True).astype(np.float64, copy=False)
        # fmt: on

    # TODO(michalk8): remove the docrep in 2.0
    @inject_docs(m=VelocityModel, b=BackwardMode, s=Similarity)  # don't swap the order
    @d.dedent
    def compute_transition_matrix(
        self,
        model: Literal["deterministic", "stochastic", "monte_carlo"] = VelocityModel.DETERMINISTIC,
        backward_mode: Literal["transpose", "negate"] = BackwardMode.TRANSPOSE,
        similarity: Union[
            Literal["correlation", "cosine", "dot_product"],
            Callable[[np.ndarray, np.ndarray, float], Tuple[np.ndarray, np.ndarray]],
        ] = Similarity.CORRELATION,
        softmax_scale: Optional[float] = None,
        n_samples: int = 1000,
        seed: Optional[int] = None,
        **kwargs: Any,
    ) -> "VelocityKernel":
        """Compute transition matrix based on velocity directions on the local manifold.

        For each cell, infer transition probabilities based on the cell's velocity-extrapolated cell state and the
        cell states of its *K* nearest neighbors.

        Parameters
        ----------
        %(velocity_mode)s
        %(velocity_backward_mode)s
        %(softmax_scale)s
        %(velocity_scheme)s
        n_samples
            Number of samples when ``mode = {m.MONTE_CARLO!r}``.
        seed
            Random seed when ``mode = {m.MONTE_CARLO!r}``.
        %(parallel)s
        kwargs
            Keyword arguments for the underlying ``model``.

        Returns
        -------
        Returns self and updates :attr:`transition_matrix`, :attr:`logits` and :attr:`params`.
        """
        start = logg.info(f"Computing transition matrix using `{model!r}` model")

        # fmt: off
        params = {"model": str(model), "similarity": str(similarity), "softmax_scale": softmax_scale}
        if self.backward:
            params["bwd_mode"] = str(backward_mode)
        if VelocityModel(model) == VelocityModel.MONTE_CARLO:
            params["n_samples"] = n_samples
            params["seed"] = seed
        if self._reuse_cache(params, time=start):
            return self

        if softmax_scale is None:
            softmax_scale = self._estimate_softmax_scale(backward_mode=backward_mode, similarity=similarity)
            logg.info(f"Using `softmax_scale={softmax_scale:.4f}`")
            params["softmax_scale"] = softmax_scale
        # fmt: on

        model = self._create_model(
            model,
            backward_mode=backward_mode,
            similarity=similarity,
            softmax_scale=softmax_scale,
            n_samples=n_samples,
            seed=seed,
        )
        if isinstance(model, Stochastic):
            kwargs["backend"] = DEFAULT_BACKEND
        self.transition_matrix, self._logits = model(**kwargs)

        logg.info("    Finish", time=start)

        return self

    def _create_model(
        self,
        model: Union[str, VelocityModel],
        backward_mode: Literal["negate", "transpose"],
        similarity: Union[str, Similarity, Callable],
        n_samples: int = 1000,
        seed: Optional[int] = None,
        **kwargs: Any,
    ):
        model = VelocityModel(model)
        backward_mode = BackwardMode(backward_mode) if self.backward else None

        if isinstance(similarity, str):
            similarity = SimilarityABC.create(Similarity(similarity))
        elif not callable(similarity):
            raise TypeError(f"Expected `scheme` to be a function, found `{type(similarity).__name__}`.")

        if self.backward and model != VelocityModel.DETERMINISTIC:
            logg.warning(
                f"Mode `{model!r}` is currently not supported for backward process. "
                f"Using to `mode={VelocityModel.DETERMINISTIC!r}`"
            )
            model = VelocityModel.DETERMINISTIC

        if model == VelocityModel.STOCHASTIC and not hasattr(similarity, "hessian"):
            model = VelocityModel.MONTE_CARLO
            logg.warning(
                f"Unable to detect a method for Hessian computation. If using one of the "
                f"predefined similarity functions, consider installing `jax` as "
                f"`pip install jax`. Using `mode={model!r}` and `n_samples={n_samples}`"
            )

        if model == VelocityModel.DETERMINISTIC:
            return Deterministic(
                self.connectivities,
                self._xdata,
                self._vdata,
                similarity=similarity,
                backward_mode=backward_mode,
                **kwargs,
            )
        if model == VelocityModel.STOCHASTIC:
            return Stochastic(
                self.connectivities,
                self._xdata,
                self._vexp,
                self._vvar,
                similarity=similarity,
                backward_mode=backward_mode,
                **kwargs,
            )
        if model == VelocityModel.MONTE_CARLO:
            return MonteCarlo(
                self.connectivities,
                self._xdata,
                self._vexp,
                self._vvar,
                similarity=similarity,
                backward_mode=backward_mode,
                n_samples=n_samples,
                seed=seed,
                **kwargs,
            )

        raise NotImplementedError(f"Model `{model}` is not yet implemented.")

    def _estimate_softmax_scale(
        self,
        n_jobs: Optional[int] = None,
        backend: Backend_t = DEFAULT_BACKEND,
        **kwargs,
    ) -> float:
        model = self._create_model(VelocityModel.DETERMINISTIC, softmax_scale=1.0, **kwargs)
        _, logits = model(n_jobs, backend)
        return 1.0 / np.median(np.abs(logits.data))

    def _extract_data(
        self,
        key: Optional[str] = None,
        attr: Optional[Union[None, Literal["layers", "obsm"]]] = None,
        subset: Optional[Union[str, Sequence[str]]] = None,
        dtype: np.dtype = np.float64,
    ) -> np.ndarray:
        if key in (None, "X"):
            data = self.adata.X
        elif attr == "layers" and key in self.adata.layers:
            data = self.adata.layers[key]
        elif attr == "obsm" and key in self.adata.obsm:
            data = np.asarray(self.adata.obsm[key])
        else:
            raise KeyError(f"Unable to find data in `adata.{attr}[{key!r}]`.")

        if subset is not None:
            if isinstance(subset, str):
                subset = self.adata.var[subset]
            subset = np.asarray(subset)
            if np.issubdtype(subset.dtype, bool) and subset.shape == (data.shape[1],):
                data = data[:, subset]
            else:
                data = data[:, np.isin(self.adata.var_names, subset)]

        data = data.astype(dtype, copy=False)
        return data.toarray() if sp.issparse(data) else data

    @property
    def logits(self) -> Optional[np.ndarray]:
        """Array of shape ``(n_cells, n_cells)`` containing the unnormalized transition matrix."""
        return self._logits

    def __invert__(self) -> "VelocityKernel":
        dk = self._copy_ignore("_transition_matrix", "_logits")
        dk._backward = not self.backward
        dk._params = {}
        return dk



from typing import Any, Callable, Optional

__all__ = ["Key"]


class cprop:
    """Class property."""

    def __init__(self, f: Callable[..., str]):
        self.f = f

    def __get__(self, obj: Any, owner: Any) -> str:
        return self.f(owner)


class Key:
    """Class which manages keys in :class:`anndata.AnnData`."""

    @classmethod
    def backward(cls, bwd: Optional[bool]) -> str:
        return "bwd" if bwd else "fwd"

    @classmethod
    def where(cls, bwd: Optional[bool]) -> str:
        return "from" if bwd else "to"

    @classmethod
    def initial(cls, bwd: Optional[bool]) -> str:
        return "initial" if bwd else "terminal"

    @classmethod
    def cytotrace(cls, key: str) -> str:
        return f"ct_{key}"

    class obs:
        @classmethod
        def probs(cls, key: str) -> str:
            return f"{key}_probs"

        @classmethod
        def macrostates(cls, bwd: Optional[bool]) -> str:
            return f"macrostates_{Key.backward(bwd)}"

        @classmethod
        def term_states(cls, estim_bwd: Optional[bool], *, bwd: bool = False) -> str:
            states = "init_states" if bwd else "term_states"
            return f"{states}_{Key.backward(estim_bwd)}"

        @classmethod
        def priming_degree(cls, bwd: Optional[bool]) -> str:
            return f"priming_degree_{Key.backward(bwd)}"

    class obsm:
        @classmethod
        def memberships(cls, key: str) -> str:
            return f"{key}_memberships"

        @classmethod
        def schur_vectors(cls, bwd: Optional[bool]) -> str:
            return f"schur_vectors_{Key.backward(bwd)}"

        @classmethod
        def macrostates(cls, bwd: Optional[bool]) -> str:
            return f"macrostates_{Key.backward(bwd)}"

        @classmethod
        def fate_probs(cls, bwd: Optional[bool]) -> str:
            return f"lineages_{Key.backward(bwd)}"

        @classmethod
        def abs_times(cls, bwd: Optional[bool]) -> str:
            return f"absorption_times_{Key.backward(bwd)}"

    class varm:
        @classmethod
        def lineage_drivers(cls, bwd: Optional[bool]):
            return Key.initial(bwd) + "_lineage_drivers"

    class uns:
        @classmethod
        def kernel(cls, bwd: Optional[bool], key: Optional[str] = None) -> str:
            return key if key is not None else f"T_{Key.backward(bwd)}"

        @classmethod
        def estimator(cls, bwd: Optional[bool], key: Optional[str] = None) -> str:
            return key if key is not None else f"{Key.backward(bwd)}_estimator"

        @classmethod
        def names(cls, key: str) -> str:
            return f"{key}_names"

        @classmethod
        def colors(cls, key: str) -> str:
            return f"{key}_colors"

        @classmethod
        def eigen(cls, bwd: Optional[bool]) -> str:
            return f"eigendecomposition_{Key.backward(bwd)}"

        @classmethod
        def schur_matrix(cls, bwd: Optional[bool]) -> str:
            return f"schur_matrix_{Key.backward(bwd)}"

        @classmethod
        def coarse(cls, bwd: Optional[bool]) -> str:
            return f"coarse_{Key.backward(bwd)}"



import enum
from typing import Any, Callable, Literal, Optional, Union

import numpy as np

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._enum import DEFAULT_BACKEND, Backend_t, ModeEnum
from cellrank._utils._utils import _connected, _irreducible
from cellrank.kernels._base_kernel import BidirectionalKernel
from cellrank.kernels.mixins import ConnectivityMixin
from cellrank.kernels.utils._pseudotime_scheme import (
    CustomThresholdScheme,
    HardThresholdScheme,
    SoftThresholdScheme,
    ThresholdSchemeABC,
)

__all__ = ["PseudotimeKernel"]


class ThresholdScheme(ModeEnum):
    SOFT = enum.auto()
    HARD = enum.auto()


@d.dedent
class PseudotimeKernel(ConnectivityMixin, BidirectionalKernel):
    """Kernel which computes directed transition probabilities based on a k-NN graph and pseudotime.

    .. seealso::
        - See :doc:`../../../notebooks/tutorials/kernels/300_pseudotime` on how to
          compute the :attr:`~cellrank.kernels.PseudotimeKernel.transition_matrix` based on the pseudotime.

    The k-NN graph contains information about the (undirected) connectivities among cells, reflecting their similarity.
    Pseudotime can be used to either remove edges that point against the direction of increasing pseudotime
    :cite:`setty:19` or to down-weight them :cite:`stassen:21`.

    Parameters
    ----------
    %(adata)s
    %(backward)s
        If :obj:`True`, the :attr:`pseudotime` will be set to ``max(pseudotime) - pseudotime``.
    time_key
        Key in :attr:`~anndata.AnnData.obs` where the pseudotime is stored.
    kwargs
        Keyword arguments for the :class:`~cellrank.kernels.ConnectivityKernel`.
    """

    def __init__(
        self,
        adata: AnnData,
        time_key: str,
        backward: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            adata,
            backward=backward,
            time_key=time_key,
            **kwargs,
        )

    def _read_from_adata(self, time_key: str, **kwargs: Any) -> None:
        super()._read_from_adata(**kwargs)
        # fmt: off
        self._time_key = time_key
        if time_key not in self.adata.obs:
            raise KeyError(f"Unable to find pseudotime in `adata.obs[{time_key!r}]`.")

        self._pseudotime = np.array(self.adata.obs[time_key]).astype(np.float64, copy=True)
        if np.any(np.isnan(self._pseudotime)):
            raise ValueError("Encountered NaN values in pseudotime.")
        # fmt: on

    @d.dedent
    def compute_transition_matrix(
        self,
        threshold_scheme: Union[
            Literal["soft", "hard"],
            Callable[[float, np.ndarray, np.ndarray], np.ndarray],
        ] = "hard",
        frac_to_keep: float = 0.3,
        b: float = 10.0,
        nu: float = 0.5,
        check_irreducibility: bool = False,
        n_jobs: Optional[int] = None,
        backend: Backend_t = DEFAULT_BACKEND,
        show_progress_bar: bool = True,
        **kwargs: Any,
    ) -> "PseudotimeKernel":
        """Compute transition matrix based on k-NN graph and pseudotemporal ordering.

        Depending on the choice of the ``threshold_scheme``, it is based on ideas by either *Palantir*
        :cite:`setty:19` or *VIA* :cite:`stassen:21`.

        Parameters
        ----------
        threshold_scheme
            Which method to use when biasing the graph. Valid options are:

            - ``'hard'`` - based on *Palantir* :cite:`setty:19` which removes some edges that point against
              the direction of increasing pseudotime. To avoid disconnecting the graph, it does not
              remove all edges that point against the direction of increasing pseudotime, but keeps the ones
              that point to cells inside a close radius. This radius is chosen according to the local cell density.
            - ``'soft'`` - based on *VIA* :cite:`stassen:21` which down-weights edges that points against
              the direction of increasing pseudotime. Essentially, the further "behind"
              a query cell is in pseudotime with respect
              to the current reference cell, the more penalized will be its graph-connectivity.
            - :class:`callable` - any function conforming to the signature of
              :func:`cellrank.kernels.utils.ThresholdSchemeABC.__call__`.
        frac_to_keep
            Fraction of the closest neighbors (according to graph connectivities) are kept, no matter whether they lie
            in the pseudotemporal past or future. This is done to ensure that the graph remains connected.
            Only used when ``threshold_scheme = 'hard'``. Must be in :math:`[0, 1]`.
        %(soft_scheme_kernel)s
        check_irreducibility
            Optional check for irreducibility of the final transition matrix.
        %(parallel)s
        kwargs
            Keyword arguments for ``threshold_scheme``.

        Returns
        -------
        Returns self and updates :attr:`transition_matrix` and :attr:`params`.
        """
        if self.pseudotime is None:
            raise ValueError("Compute pseudotime first.")  # CytoTraceKernel

        start = logg.info("Computing transition matrix based on pseudotime")
        if isinstance(threshold_scheme, str):
            threshold_scheme = ThresholdScheme(threshold_scheme)
            if threshold_scheme == ThresholdScheme.SOFT:
                scheme = SoftThresholdScheme()
                kwargs["b"] = b
                kwargs["nu"] = nu
            elif threshold_scheme == ThresholdScheme.HARD:
                scheme = HardThresholdScheme()
                kwargs["frac_to_keep"] = frac_to_keep
            else:
                raise NotImplementedError(f"Threshold scheme `{threshold_scheme}` is not yet implemented.")
        elif isinstance(threshold_scheme, ThresholdSchemeABC):
            scheme = threshold_scheme
        elif callable(threshold_scheme):
            scheme = CustomThresholdScheme(threshold_scheme)
        else:
            raise TypeError(
                f"Expected `threshold_scheme` to be either a `str` or a `callable`, found `{type(threshold_scheme)}`."
            )

        # fmt: off
        if self._reuse_cache({"dnorm": False, "scheme": str(threshold_scheme), **kwargs}, time=start):
            return self
        # fmt: on

        biased_conn = scheme.bias_knn(
            self.connectivities,
            self.pseudotime,
            n_jobs=n_jobs,
            backend=backend,
            show_progress_bar=show_progress_bar,
            **kwargs,
        )

        # make sure the biased graph is still connected
        if not _connected(biased_conn):
            logg.warning("Biased k-NN graph is disconnected")
        if check_irreducibility and not _irreducible(biased_conn):
            logg.warning("Biased k-NN graph is not irreducible")

        self.transition_matrix = biased_conn
        logg.info("    Finish", time=start)

        return self

    @property
    def pseudotime(self) -> Optional[np.array]:
        """Pseudotemporal ordering of cells."""
        if self._pseudotime is None:
            return None
        if self.backward:
            return np.max(self._pseudotime) - self._pseudotime
        return self._pseudotime

    def __invert__(self) -> "PseudotimeKernel":
        pk = self._copy_ignore("_transition_matrix")
        pk._backward = not self.backward
        pk._params = {}
        return pk



import copy
import enum
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from wrapt import decorator

import numpy as np
import pandas as pd

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._enum import ModeEnum
from cellrank._utils._lineage import Lineage

__all__ = [
    "BaseProtocol",
    "logger",
    "shadow",
    "SafeGetter",
    "StatesHolder",
]


class StatesHolder(NamedTuple):  # noqa: D101
    assignment: Optional[pd.Series] = None
    probs: Optional[pd.Series] = None
    colors: Optional[np.ndarray] = None
    memberships: Optional[Lineage] = None

    def set(self, **kwargs: Any) -> "StatesHolder":  # noqa: D102
        return self._replace(**kwargs)


class PlotMode(ModeEnum):
    EMBEDDING = enum.auto()
    TIME = enum.auto()


class BaseProtocol(Protocol):  # noqa: D101
    @property
    def adata(self) -> AnnData:  # noqa: D102
        ...

    @property
    def backward(self) -> bool:  # noqa: D102
        ...

    @property
    def params(self) -> Dict[str, Any]:  # noqa: D102
        ...

    def _set(
        self,
        attr: Optional[str] = None,
        obj: Optional[Union[pd.DataFrame, Mapping[str, Any]]] = None,
        key: Optional[str] = None,
        value: Optional[Union[np.ndarray, pd.Series, pd.DataFrame, Lineage, AnnData, Dict[str, Any]]] = None,
        copy: bool = True,
        shadow_only: bool = False,
    ) -> None:
        ...

    def _get(
        self,
        *,
        obj: Union[pd.DataFrame, Mapping[str, Any]],
        key: str,
        shadow_attr: Optional[Literal["obs", "obsm", "var", "varm", "uns"]] = None,
        dtype: Optional[Union[type, Tuple[type, ...]]] = None,
        copy: bool = True,
        allow_missing: bool = False,
    ) -> Any:
        ...

    def _create_params(
        self,
        locs: Optional[Mapping[str, Any]] = None,
        func: Optional[Callable] = None,
        remove: Sequence[str] = (),
    ) -> Dict[str, Any]:
        ...

    def _read_params(self, key: str) -> Dict[str, Any]:
        ...


@decorator()
def logger(wrapped: Callable[..., str], instance: Any, args: Any, kwargs: Dict[str, Any]) -> str:
    """Handle logging for :class:`~anndata.AnnData` writing functions of :class:`~cellrank.estimators.BaseEstimator`."""
    log, time = kwargs.pop("log", True), kwargs.pop("time", None)
    msg = wrapped(*args, **kwargs)

    if log:
        logg.info(msg, time=time)

    return msg


@decorator()
def shadow(wrapped: Callable[..., str], instance: Any, args: Any, kwargs: Mapping[str, Any]) -> str:
    """Duplicate function call with shadow :class:`~anndata.AnnData` object.

    Used to create :class:`~anndata.AnnData` serialization object in the :class:`~cellrank.estimators.BaseEstimator`.
    """
    res = wrapped(*args, **kwargs)
    with instance._shadow:
        try:
            # don't care what the "shadowed" function returns
            _ = wrapped(*args, **kwargs)
        except Exception as e:  # noqa: BLE001
            logg.error(f"Unable to duplicate function call using shadow `anndata.AnnData` object. Reason: `{e}`")

    return res


class SafeGetter:
    """Context manager that does a simple rollback on :attr:`object.__dict__`.

    Used in :class:`~cellrank.estimators.BaseEstimator` to maintain consistency
    during deserialization from :class:`~anndata.AnnData`.

    Parameters
    ----------
    obj
        Object to watch. Shallow copy of :attr:`obj.__dict__` is made during initialization,
        except `_shadow_adata` (deep copy), if present.
    allowed
        Whether to allow these exception to occur with the context.

    Notes
    -----
    :attr:`ok` is :obj:`False` if and only if any exception, regardless of whether it is in ``allowed``,
    has been encountered.
    """

    def __init__(self, obj: Any, allowed: Union[type, Sequence[type]] = ()):
        self._exc_type: Optional[TypeVar[Exception]] = None
        self._obj = obj
        if not isinstance(allowed, Iterable):
            allowed = (allowed,)
        self._allowed = allowed

        self._dict = copy.copy(obj.__dict__)
        if "_shadow_adata" in self._dict:
            # deepcopy `_shadow_adata` (small cost) to easily revert changes
            # do not run deepcopy on the full `__dict__` (possibly large cost)
            # silent assumption is that within the context `adata` is not modified
            self._dict["_shadow_adata"] = copy.deepcopy(self._dict["_shadow_adata"])

    def __enter__(self) -> "SafeGetter":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        self._exc_type = exc_type
        # TODO(michalk8): log properly
        if not self.ok:
            logg.debug(f"The estimator will not be completely initialized, reason: {exc_type(exc_val)}")
            self._obj.__dict__ = self._dict
            return self._exc_type in self._allowed

        return True

    @property
    def ok(self) -> bool:
        """Return :obj:`True` if no exception has been raised."""
        return self._exc_type is None



from cellrank.kernels.mixins._anndata import AnnDataMixin
from cellrank.kernels.mixins._io import IOMixin
from cellrank.kernels.mixins._kernel import (
    BidirectionalMixin,
    ConnectivityMixin,
    UnidirectionalMixin,
)

__all__ = ["AnnDataMixin", "IOMixin", "BidirectionalMixin", "ConnectivityMixin", "UnidirectionalMixin"]



import pathlib
import types
from typing import Any, Dict, Literal, Mapping, Optional, Tuple, Union

import numpy as np
import scipy.sparse as sp

import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter, MultipleLocator

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._key import Key
from cellrank._utils._utils import _eigengap, save_fig
from cellrank.estimators.mixins._utils import BaseProtocol, SafeGetter, logger, shadow

__all__ = ["EigenMixin"]

EPS = np.finfo(np.float64).eps


class EigenProtocol(BaseProtocol):
    @property
    def transition_matrix(self) -> Union[np.ndarray, sp.spmatrix]:
        ...

    def _write_eigendecomposition(
        self, decomp: Dict[str, Any], params: Mapping[str, Any] = types.MappingProxyType({})
    ) -> str:
        ...


class EigenMixin:
    """Mixin that computes Schur decomposition."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._eigendecomposition = None

    @property
    @d.get_summary(base="eigen")
    def eigendecomposition(self) -> Optional[Dict[str, Any]]:
        """Eigendecomposition of the :attr:`transition_matrix`.

        For non-symmetric real matrices, left and right eigenvectors will in general be different and complex. We
        compute both left and right eigenvectors.

        Returns
        -------
        A dictionary with the following keys:

        - ``'D'`` - the eigenvalues.
        - ``'eigengap'`` - the `eigengap <https://en.wikipedia.org/wiki/Eigengap>`__.
        - ``'params'`` - parameters used for the computation.
        - ``'V_l'`` - left eigenvectors (optional).
        - ``'V_r'`` - right eigenvectors (optional).
        - ``'stationary_dist'`` - stationary distribution of the :attr:`transition_matrix`, if present.
        """
        return self._eigendecomposition

    @d.dedent
    def compute_eigendecomposition(
        self: EigenProtocol,
        k: int = 20,
        which: Literal["LR", "LM"] = "LR",
        alpha: float = 1.0,
        only_evals: bool = False,
        ncv: Optional[int] = None,
    ) -> "EigenMixin":
        """Compute eigendecomposition of the :attr:`transition_matrix`.

        Uses a sparse implementation, if possible, and only computes the top :math:`k` eigenvectors
        to speed up the computation. Computes both left and right eigenvectors.

        Parameters
        ----------
        k
            Number of eigenvectors or eigenvalues to compute.
        %(eigen)s
        only_evals
            Whether to compute only eigenvalues.
        ncv
            Number of Lanczos vectors generated.

        Returns
        -------
        Self and updates the following fields:

        - :attr:`eigendecomposition` - %(eigen.summary)s
        """

        def get_top_k_evals() -> np.ndarray:
            return D[np.flip(np.argsort(D.real))][:k]

        start = logg.info("Computing eigendecomposition of the transition matrix")

        if sp.issparse(self.transition_matrix):
            logg.debug(f"Computing top `{k}` eigenvalues of a sparse matrix")
            D, V_l = sp.linalg.eigs(self.transition_matrix.T, k=k, which=which, ncv=ncv)
            if only_evals:
                self._write_eigendecomposition(
                    {
                        "D": get_top_k_evals(),
                        "eigengap": _eigengap(get_top_k_evals().real, alpha),
                        "params": {"which": which, "k": k, "alpha": alpha},
                    },
                    time=start,
                )
                return
            _, V_r = sp.linalg.eigs(self.transition_matrix, k=k, which=which, ncv=ncv)
        else:
            logg.warning("This transition matrix is not sparse, computing full eigendecomposition")
            D, V_l = np.linalg.eig(self.transition_matrix.T)
            if only_evals:
                self._write_eigendecomposition(
                    {
                        "D": get_top_k_evals(),
                        "eigengap": _eigengap(D.real, alpha),
                        "params": {"which": which, "k": k, "alpha": alpha},
                    },
                    time=start,
                )
                return
            _, V_r = np.linalg.eig(self.transition_matrix)

        # Sort the eigenvalues and eigenvectors and take the real part
        logg.debug("Sorting eigenvalues by their real part")
        p = np.flip(np.argsort(D.real))
        D, V_l, V_r = D[p], V_l[:, p], V_r[:, p]
        e_gap = _eigengap(D.real, alpha)

        pi = np.abs(V_l[:, 0].real)
        pi /= np.sum(pi)

        self._write_eigendecomposition(
            {
                "D": D,
                "stationary_dist": pi,
                "V_l": V_l,
                "V_r": V_r,
                "eigengap": e_gap,
                "params": {"which": which, "k": k, "alpha": alpha},
            },
            time=start,
        )
        return self

    @d.dedent
    def plot_spectrum(
        self,
        n: Optional[int] = None,
        real_only: Optional[bool] = None,
        show_eigengap: bool = True,
        show_all_xticks: bool = True,
        legend_loc: Optional[str] = None,
        title: Optional[str] = None,
        marker: str = ".",
        figsize: Optional[Tuple[float, float]] = (5, 5),
        dpi: int = 100,
        save: Optional[Union[str, pathlib.Path]] = None,
        **kwargs: Any,
    ) -> None:
        """Plot the top eigenvalues in a real or a complex plane.

        Parameters
        ----------
        n
            Number of eigenvalues to show. If :obj:`None`, show all that have been computed.
        real_only
            Whether to plot only the real part of the spectrum.
            If :obj:`None`, plot real spectrum if no complex eigenvalues are present.
        show_eigengap
            When ``real_only = True``, this determines whether to show the inferred eigengap as a dotted line.
        show_all_xticks
            When ``real_only = True``, this determines whether to show the indices of all eigenvalues on the x-axis.
        legend_loc
            Location parameter for the legend.
        title
            Title of the figure.
        marker
            Marker symbol used, valid options can be found in :mod:`~matplotlib.markers`.
        %(plotting)s
        kwargs
            Keyword arguments for :meth:`~matplotlib.axes.Axes.scatter`.

        Returns
        -------
        %(just_plots)s
        """
        eig = self.eigendecomposition
        if eig is None:
            raise RuntimeError("Compute eigendecomposition first as `.compute_eigendecomposition()`.")

        if n is None:
            n = len(eig["D"])
        if n <= 0:
            raise ValueError(f"Expected `n` to be > 0, found `{n}`.")
        if real_only is None:
            real_only = np.any(np.iscomplex(eig["D"][:n]))

        if real_only:
            fig = self._plot_real_spectrum(
                n,
                show_eigengap=show_eigengap,
                show_all_xticks=show_all_xticks,
                dpi=dpi,
                figsize=figsize,
                legend_loc=legend_loc,
                title=title,
                marker=marker,
                **kwargs,
            )
        else:
            fig = self._plot_complex_spectrum(
                n,
                dpi=dpi,
                figsize=figsize,
                legend_loc=legend_loc,
                title=title,
                marker=marker,
                **kwargs,
            )

        if save:
            save_fig(fig, save)

    def _plot_complex_spectrum(
        self,
        n: int,
        dpi: int = 100,
        figsize: Optional[Tuple[float, float]] = None,
        legend_loc: Optional[str] = None,
        title: Optional[str] = None,
        marker: str = ".",
        **kwargs: Any,
    ) -> plt.Figure:
        # define a function to make the data limits rectangular
        def adapt_range(min_: float, max_: float, range_: float) -> Tuple[float, float]:
            return (
                min_ + (max_ - min_) / 2 - range_ / 2,
                min_ + (max_ - min_) / 2 + range_ / 2,
            )

        eig = self.eigendecomposition
        D, params = eig["D"][:n], eig["params"]

        # create figure and axes
        fig, ax = plt.subplots(nrows=1, ncols=1, dpi=dpi, figsize=figsize)

        # get the original data ranges
        lam_x, lam_y = D.real, D.imag
        x_min, x_max = np.min(lam_x), np.max(lam_x)
        y_min, y_max = np.min(lam_y), np.max(lam_y)
        x_range, y_range = x_max - x_min, y_max - y_min
        final_range = np.max([x_range, y_range]) + 0.05

        x_min_, x_max_ = adapt_range(x_min, x_max, final_range)
        y_min_, y_max_ = adapt_range(y_min, y_max, final_range)

        # plot the data and the unit circle
        ax.scatter(D.real, D.imag, marker=marker, label="eigenvalue", **kwargs)
        t = np.linspace(0, 2 * np.pi, 500)
        x_circle, y_circle = np.sin(t), np.cos(t)
        ax.plot(x_circle, y_circle, "k-", label="unit circle")

        # set labels, ranges and legend
        ax.set_xlabel(r"Re($\lambda$)")
        ax.set_xlim(x_min_, x_max_)

        ax.set_ylabel(r"Im($\lambda$)")
        ax.set_ylim(y_min_, y_max_)

        key = "real part" if params["which"] == "LR" else "magnitude"
        if title is None:
            title = f"top {n} eigenvalues according to their {key}"

        ax.set_title(title)
        if legend_loc not in (None, "none"):
            ax.legend(loc=legend_loc)

        return fig

    def _plot_real_spectrum(
        self,
        n: int,
        show_eigengap: bool = True,
        show_all_xticks: bool = True,
        dpi: int = 100,
        figsize: Optional[Tuple[float, float]] = None,
        legend_loc: Optional[str] = None,
        title: Optional[str] = None,
        marker: str = ".",
        **kwargs: Any,
    ) -> plt.Figure:
        eig = self.eigendecomposition
        D, params = eig["D"][:n], eig["params"]

        D_real, D_imag = D.real, D.imag
        ixs = np.arange(len(D))
        mask = D_imag == 0

        # plot the top eigenvalues
        fig, ax = plt.subplots(nrows=1, ncols=1, dpi=dpi, figsize=figsize)
        if np.any(mask):
            ax.scatter(
                ixs[mask],
                D_real[mask],
                marker=marker,
                label="real eigenvalue",
                **kwargs,
            )
        if np.any(~mask):
            ax.scatter(
                ixs[~mask],
                D_real[~mask],
                marker=marker,
                label="complex eigenvalue",
                **kwargs,
            )

        # add dashed line for the eigengap, ticks, labels, title and legend
        if show_eigengap and eig["eigengap"] < n:
            ax.axvline(eig["eigengap"], label="eigengap", ls="--", lw=2, c="k")

        ax.set_xlabel("index")
        if show_all_xticks:
            ax.set_xticks(np.arange(len(D)))
            ax.set_xticklabels(np.arange(1, len(D) + 1))
        else:
            ax.xaxis.set_major_locator(MultipleLocator(2.0))
            ax.xaxis.set_major_formatter(FormatStrFormatter("%d"))

        ax.set_ylabel(r"Re($\lambda_i$)")

        key = "real part" if params["which"] == "LR" else "magnitude"
        if title is None:
            title = f"real part of top {n} eigenvalues according to their {key}"

        ax.set_title(title)
        if legend_loc not in (None, "none"):
            ax.legend(loc=legend_loc)

        return fig

    @logger
    @shadow
    def _write_eigendecomposition(
        self: EigenProtocol,
        decomp: Dict[str, Any],
        params: Mapping[str, Any] = types.MappingProxyType({}),
    ) -> str:
        key = Key.uns.eigen(self.backward)
        self._set("_eigendecomposition", self.adata.uns, key=key, value=decomp)
        params = params or decomp.get("params", {})
        self.params[key] = dict(params)

        return f"Adding `adata.uns[{key!r}]`\n" f"       `.eigendecomposition`\n" "    Finish"

    def _read_eigendecomposition(self: EigenProtocol, adata: AnnData, allow_missing: bool = True) -> bool:
        key = Key.uns.eigen(self.backward)
        with SafeGetter(self, allowed=KeyError) as sg:
            self._eigendecomposition = self._get(
                obj=adata.uns,
                key=key,
                shadow_attr="uns",
                dtype=dict,
                allow_missing=allow_missing,
            )
            self.params[key] = self._read_params(key)

        return sg.ok



from cellrank.estimators.terminal_states._cflare import CFLARE
from cellrank.estimators.terminal_states._gpcca import GPCCA
from cellrank.estimators.terminal_states._term_states_estimator import (
    TermStatesEstimator,
)

__all__ = ["CFLARE", "GPCCA", "TermStatesEstimator"]



from cellrank.kernels import utils
from cellrank.kernels._base_kernel import (
    BidirectionalKernel,
    Kernel,
    UnidirectionalKernel,
)
from cellrank.kernels._connectivity_kernel import ConnectivityKernel
from cellrank.kernels._cytotrace_kernel import CytoTRACEKernel
from cellrank.kernels._experimental_time_kernel import ExperimentalTimeKernel
from cellrank.kernels._precomputed_kernel import PrecomputedKernel
from cellrank.kernels._pseudotime_kernel import PseudotimeKernel
from cellrank.kernels._real_time_kernel import RealTimeKernel
from cellrank.kernels._velocity_kernel import VelocityKernel

__all__ = [
    "utils",
    "Kernel",
    "UnidirectionalKernel",
    "BidirectionalKernel",
    "ConnectivityKernel",
    "CytoTRACEKernel",
    "ExperimentalTimeKernel",
    "PrecomputedKernel",
    "PseudotimeKernel",
    "RealTimeKernel",
    "VelocityKernel",
]



import multiprocessing
import threading
from typing import Any, Callable, Optional, Sequence, Union

import joblib as jl

import numpy as np
import scipy.sparse as sp

__all__ = ["parallelize", "_get_n_cores"]


def parallelize(
    callback: Callable[[Any], Any],
    collection: Union[sp.spmatrix, Sequence[Any]],
    n_jobs: Optional[int] = None,
    n_split: Optional[int] = None,
    unit: str = "",
    as_array: bool = True,
    use_ixs: bool = False,
    backend: str = "loky",
    extractor: Optional[Callable[[Any], Any]] = None,
    show_progress_bar: bool = True,
) -> Any:
    """Parallelize function call over a collection of elements.

    Parameters
    ----------
    callback
        Function to parallelize.
    collection
        Sequence of items which to chunkify or an already chunked array.
    n_jobs
        Number of parallel jobs.
    n_split
        Split ``collection`` into ``n_split`` chunks. If `None`, split into ``n_jobs`` chunks.
    unit
        Unit of the progress bar.
    as_array
        Whether to convert the results not :class:`numpy.ndarray`.
    use_ixs
        Whether to pass indices to the callback.
    backend
        Which backend to use for multiprocessing. See :class:`joblib.Parallel` for valid options.
    extractor
        Function to apply to the result after all jobs have finished.
    show_progress_bar
        Whether to show a progress bar.

    Returns
    -------
    The result depending on ``callable``, ``extractor`` and ``as_array``.
    """
    if show_progress_bar:
        try:
            from tqdm.auto import tqdm
        except ImportError:
            try:
                from tqdm.std import tqdm
            except ImportError:
                tqdm = None
    else:
        tqdm = None

    def update(pbar, queue, n_total):
        n_finished = 0
        while n_finished < n_total:
            try:
                res = queue.get()
            except EOFError as e:
                if not n_finished != n_total:
                    raise RuntimeError(f"Finished only `{n_finished}` out of `{n_total}` tasks.`") from e
                break
            assert res in (None, (1, None), 1)  # (None, 1) means only 1 job
            if res == (1, None):
                n_finished += 1
                if pbar is not None:
                    pbar.update()
            elif res is None:
                n_finished += 1
            elif pbar is not None:
                pbar.update()

        if pbar is not None:
            pbar.close()

    def wrapper(*args, **kwargs):
        if pass_queue and show_progress_bar:
            pbar = None if tqdm is None else tqdm(total=col_len, unit=unit, mininterval=0.125)
            queue = multiprocessing.Manager().Queue()
            thread = threading.Thread(target=update, args=(pbar, queue, len(collections)))
            thread.start()
        else:
            pbar, queue, thread = None, None, None

        res = jl.Parallel(n_jobs=n_jobs, backend=backend)(
            jl.delayed(callback)(
                *((i, cs) if use_ixs else (cs,)),
                *args,
                **kwargs,
                queue=queue,
            )
            for i, cs in enumerate(collections)
        )

        res = np.array(res) if as_array else res
        if thread is not None:
            thread.join()

        return res if extractor is None else extractor(res)

    col_len = collection.shape[0] if sp.issparse(collection) else len(collection)
    n_jobs = _get_n_cores(n_jobs, col_len)
    if n_split is None:
        n_split = n_jobs

    if sp.issparse(collection):
        n_split = max(1, min(n_split, collection.shape[0]))
        if n_split == collection.shape[0]:
            collections = [collection[[ix], :] for ix in range(collection.shape[0])]
        else:
            step = collection.shape[0] // n_split
            ixs = [np.arange(i * step, min((i + 1) * step, collection.shape[0])) for i in range(n_split)]
            ixs[-1] = np.append(ixs[-1], np.arange(ixs[-1][-1] + 1, collection.shape[0]))

            collections = [collection[ix, :] for ix in filter(len, ixs)]
    else:
        collections = list(filter(len, np.array_split(collection, n_split)))

    n_split = len(collections)
    n_jobs = min(n_jobs, n_split)
    pass_queue = not hasattr(callback, "py_func")  # we'd be inside a numba function

    return wrapper


def _get_n_cores(n_cores: Optional[int], n_jobs: Optional[int]) -> int:
    """Make number of cores a positive integer.

    Parameters
    ----------
    n_cores
        Number of cores to use.
    n_jobs.
        Number of jobs. This is just used to determine if the collection is a singleton.
        If `1`, always returns `1`.

    Returns
    -------
    Positive integer corresponding to how many cores to use.
    """
    if n_cores == 0:
        raise ValueError("Number of cores cannot be `0`.")
    if n_jobs == 1 or n_cores is None:
        return 1
    if n_cores < 0:
        return multiprocessing.cpu_count() + 1 + n_cores

    return n_cores
